import dataclasses
import os
import pathlib
import resource
import subprocess
import sys
import threading
import typing
from enum import Enum
from time import monotonic
from typing import IO, Any, Dict, List, Optional, Union

import psutil

from rbx.utils import PathOrStr

FileLike = Union[PathOrStr, IO[bytes], int]


def _maybe_close_files(files):
    for fobj in files:
        if isinstance(fobj, int):
            try:
                os.close(fobj)
            except OSError:
                pass
        else:
            fobj.close()


def _is_pathlike(obj: Any) -> bool:
    return isinstance(obj, str) or isinstance(obj, pathlib.Path)


@dataclasses.dataclass
class ProgramIO:
    input: FileLike = subprocess.PIPE
    output: FileLike = subprocess.PIPE
    stderr: FileLike = subprocess.PIPE

    def get_file_objects(self):
        if isinstance(self.input, int):
            input_fobj = self.input
        elif _is_pathlike(self.input):
            input_fobj = pathlib.Path(typing.cast(str, self.input)).open('r')
        else:
            input_fobj = typing.cast(IO[bytes], self.input)
        if isinstance(self.output, int):
            output_fobj = self.output
        elif _is_pathlike(self.output):
            output_path = pathlib.Path(typing.cast(str, self.output))
            output_path.parent.mkdir(parents=True, exist_ok=True)
            output_fobj = output_path.open('w')
        else:
            output_fobj = typing.cast(IO[bytes], self.output)
        if isinstance(self.stderr, int):
            stderr_fobj = self.stderr
        elif _is_pathlike(self.stderr):
            stderr_path = pathlib.Path(typing.cast(str, self.stderr))
            stderr_path.parent.mkdir(parents=True, exist_ok=True)
            stderr_fobj = stderr_path.open('w')
        else:
            stderr_fobj = typing.cast(IO[bytes], self.stderr)
        return input_fobj, output_fobj, stderr_fobj


@dataclasses.dataclass
class ProgramPipes:
    input: Optional[IO[bytes]] = None
    output: Optional[IO[bytes]] = None
    stderr: Optional[IO[bytes]] = None


@dataclasses.dataclass
class ProgramParams:
    io: ProgramIO = dataclasses.field(default_factory=ProgramIO)
    chdir: Optional[pathlib.Path] = None
    time_limit: Optional[float] = None  # seconds
    wall_time_limit: Optional[float] = None  # seconds
    memory_limit: Optional[int] = None  # megabytes
    fs_limit: Optional[int] = None  # kilobytes
    env: Dict[str, str] = dataclasses.field(default_factory=dict)
    pgid: Optional[int] = None


def get_preexec_fn(params: ProgramParams):
    def preexec_fn():
        os.setpgid(0, params.pgid or 0)
        if params.time_limit is not None:
            time_limit_in_ms = int(params.time_limit * 1000)
            rlimit_cpu = int((time_limit_in_ms + 999) // 1000)
            resource.setrlimit(resource.RLIMIT_CPU, (rlimit_cpu, rlimit_cpu + 1))
        if params.fs_limit is not None:
            fs_limit = params.fs_limit * 1024  # in bytes
            resource.setrlimit(resource.RLIMIT_FSIZE, (fs_limit + 1, fs_limit * 2))
        if sys.platform != 'darwin':
            try:
                resource.setrlimit(
                    resource.RLIMIT_STACK,
                    (resource.RLIM_INFINITY, resource.RLIM_INFINITY),
                )
            except ValueError:
                pass

    return preexec_fn


def get_memory_usage(ru: resource.struct_rusage) -> int:
    """Get memory usage in bytes from resource usage statistics.

    Returns the total memory usage (RSS + shared memory segments) in bytes.

    Platform differences in ru.ru_maxrss:
    - macOS/Darwin: ru.ru_maxrss is in bytes
    - Linux: ru.ru_maxrss is in kilobytes

    This function normalizes the result to always return bytes.

    Args:
        ru: Resource usage statistics from os.wait4() or similar

    Returns:
        int: Total memory usage in bytes
    """
    if sys.platform == 'darwin':
        # On macOS, ru.ru_maxrss is already in bytes
        return ru.ru_maxrss + ru.ru_ixrss * 1024
    # On Linux, ru.ru_maxrss is in kilobytes, so convert to bytes
    return (ru.ru_maxrss + ru.ru_ixrss + ru.ru_idrss + ru.ru_isrss) * 1024


def get_cpu_time(ru: resource.struct_rusage) -> float:
    """Get CPU time in seconds from resource usage statistics.

    Returns the total CPU time (user + system) in seconds.

    Args:
        ru: Resource usage statistics from os.wait4() or similar

    Returns:
        float: Total CPU time in seconds
    """
    return ru.ru_utime + ru.ru_stime


def get_file_sizes(io: ProgramIO):
    return _get_file_size(io.output) + _get_file_size(io.stderr)


def _get_file_size(filename: Optional[FileLike]) -> int:
    if filename is None or not _is_pathlike(filename):
        return 0
    path = pathlib.Path(typing.cast(str, filename))
    if not path.is_file():
        return 0
    return path.stat().st_size


class ProgramCode(Enum):
    RE = 'RE'
    SG = 'SG'
    TO = 'TO'
    WT = 'WT'
    ML = 'ML'
    OL = 'OL'
    TE = 'TE'


@dataclasses.dataclass
class ProgramResult:
    exitcode: int
    wall_time: float
    cpu_time: float
    memory_used: int
    file_sizes: int
    program_codes: List[ProgramCode]
    killing_signal: Optional[int] = None
    alarm_msg: Optional[str] = None


class ProgramError(Exception):
    pass


class Program:
    def __init__(self, command: List[str], params: ProgramParams):
        self.command = command
        self.params = params
        self.popen: Optional[subprocess.Popen] = None
        self._files = []

        self._stop_wall_handler = threading.Event()
        self._stop_alarm_handler = threading.Event()
        self._alarm_msg = ''

        self._run()

    @property
    def pipes(self) -> ProgramPipes:
        assert self.popen is not None
        return ProgramPipes(
            input=self.popen.stdin,
            output=self.popen.stdout,
            stderr=self.popen.stderr,
        )

    @property
    def pid(self) -> int:
        assert self.popen is not None
        return self.popen.pid

    def _kill_process(self):
        if self.popen is not None:
            self.popen.kill()

    def _handle_wall(self):
        if self._stop_wall_handler.wait(self.params.wall_time_limit):
            return
        self._stop_alarm_handler.set()
        self._alarm_msg = 'wall timelimit'
        self._kill_process()

    def _handle_alarm(self):
        if self._stop_alarm_handler.wait(0.3):
            return
        try:
            process = psutil.Process(self.pid)
            if self.params.time_limit is not None:
                times = process.cpu_times()
                cpu_time = times.user + times.system
                if cpu_time > self.params.time_limit:
                    self._alarm_msg = 'timelimit'
                    self._kill_process()
            if self.params.memory_limit is not None:
                memory_info = process.memory_info()
                memory_used = memory_info.rss
                if memory_used > self.params.memory_limit * 1024 * 1024:
                    self._alarm_msg = 'memorylimit'
                    self._kill_process()
            self._stop_alarm_handler.clear()
            self._handle_alarm()
        except psutil.NoSuchProcess:
            return

    def _run(self):
        self._files = self.params.io.get_file_objects()
        try:
            self.popen = subprocess.Popen(
                self.command,
                stdin=self._files[0],
                stdout=self._files[1],
                stderr=self._files[2],
                cwd=self.params.chdir,
                env={**os.environ, **self.params.env},
                preexec_fn=get_preexec_fn(self.params),
            )
        except FileNotFoundError as e:
            raise ProgramError(f'Command {self.command[0]} not found') from e
        except PermissionError as e:
            raise ProgramError(f'Permission denied for command {self.command}') from e
        self.start_time = monotonic()

        threading.Thread(target=self._handle_wall, daemon=True).start()
        threading.Thread(target=self._handle_alarm, daemon=True).start()

    def close_pipes(self):
        if self.popen is None:
            return
        if self.params.io.input == subprocess.PIPE and self.pipes.input is not None:
            self.pipes.input.close()
        if self.params.io.output == subprocess.PIPE and self.pipes.output is not None:
            self.pipes.output.close()
        if self.params.io.stderr == subprocess.PIPE and self.pipes.stderr is not None:
            self.pipes.stderr.close()

    def close(self):
        self.close_pipes()
        _maybe_close_files(self._files)

    def process_exit(self, exitstatus, ru) -> ProgramResult:
        _maybe_close_files(self._files)

        wall_time = monotonic() - self.start_time
        cpu_time = get_cpu_time(ru)
        memory_used = get_memory_usage(ru)
        file_sizes = get_file_sizes(self.params.io)
        exitcode = os.waitstatus_to_exitcode(exitstatus)
        killing_signal = None
        program_codes = []

        if exitcode < 0:
            killing_signal = -exitcode
            program_codes.append(ProgramCode.SG)
        if exitcode > 0:
            program_codes.append(ProgramCode.RE)
        if self.params.time_limit is not None and (
            cpu_time > self.params.time_limit or -exitcode == 24
        ):
            program_codes.append(ProgramCode.TO)
        if (
            self.params.wall_time_limit is not None
            and wall_time > self.params.wall_time_limit
        ):
            program_codes.append(ProgramCode.WT)
            program_codes.append(ProgramCode.TO)
        # Memory limit checking: Two ways a process can exceed memory limits:
        # 1. Runtime monitoring (_handle_alarm) kills the process during execution
        # 2. Post-execution check using ru.ru_maxrss detects peak memory usage exceeded limit
        # Both memory_used (from ru.ru_maxrss) and memory_limit (converted to bytes) are in bytes
        if (
            self.params.memory_limit is not None
            and memory_used > self.params.memory_limit * 1024 * 1024
            or self._alarm_msg == 'memorylimit'
        ):
            program_codes.append(ProgramCode.ML)
        if (
            self.params.fs_limit is not None
            and file_sizes > self.params.fs_limit * 1024
        ):
            program_codes.append(ProgramCode.OL)

        result = ProgramResult(
            exitcode=exitcode,
            wall_time=wall_time,
            cpu_time=cpu_time,
            memory_used=memory_used,
            file_sizes=file_sizes,
            program_codes=program_codes,
            killing_signal=killing_signal,
            alarm_msg=self._alarm_msg or None,
        )
        return result

    def wait(self):
        assert self.popen is not None
        _, exitstatus, ru = os.wait4(self.pid, 0)
        res = self.process_exit(exitstatus, ru)
        self.close_pipes()
        return res



import ast
import pathlib
from typing import List, Optional

import lark

from rbx.box.generation_schema import GenerationInput, GeneratorScriptEntry
from rbx.box.schema import GeneratorCall, Testcase
from rbx.box.stressing import whitespace


class ScriptGeneratedInput(GenerationInput):
    """Input generated from a generator script with optional group annotation."""

    group: Optional[str] = None


LARK_GRAMMAR = r'''
start: _statement*

_statement: comment
          | copy_test
          | inline_input
          | testgroup
          | generator_call

// Comments (whole line only)
comment: COMMENT

// Generator call - name must be first token on line, consumes rest of line as args
generator_call: REST_OF_LINE

// Copy test
copy_test: COPY_KEYWORD REST_OF_LINE

// Inline input
inline_input: INPUT_KEYWORD string
            | INPUT_KEYWORD _LBRACE input_lines _RBRACE

string: TRIPLE_QUOTED_STRING | ESCAPED_STRING

// Content between braces
input_lines: BLOCK_CONTENT?

// Testgroup
testgroup: TESTGROUP_KEYWORD GROUP_NAME _LBRACE _statement* _RBRACE

// Tokens
COMMENT.3: /(\/\/|#)[^\n\r]*/
COPY_KEYWORD.3: "@copy"
INPUT_KEYWORD.3: "@input"
TESTGROUP_KEYWORD.3: "@testgroup"
# no ambiguity
REST_OF_LINE.2: /(?!(?:@copy|@input|@testgroup)\b)[^\s\/#][^\n\r]*/
FILEPATH: /[A-Za-z0-9\.][\/A-Za-z0-9\-_\.]*/
GROUP_NAME: /[a-zA-Z0-9][a-zA-Z0-9\-_]*/

// String literals - support both single and double quotes with escape sequences
// Negative lookahead (?!") prevents matching "" when it's part of """
ESCAPED_STRING: /'(?:[^'\\]|\\.)*'/ | /"(?:[^"\\]|\\.)*"(?!")/
// Triple-quoted strings (multiline)
TRIPLE_QUOTED_STRING: /"""[\s\S]*?"""/
// Content between braces
BLOCK_CONTENT: /[^\}]+/

_INDENT: /[ \t]+/
_S: /[ \t]+/
_WS: /\s+/
_LBRACE: "{"
_RBRACE: "}"
_NEWLINE: /\r?\n/

%ignore _WS
'''

LARK_PARSER = lark.Lark(LARK_GRAMMAR, propagate_positions=True)


@lark.v_args(inline=True, meta=True)
class TestPlanTransformer(lark.Transformer):
    """
    Transformer for parsing test plan files.

    Transforms the parse tree into a list of ScriptGeneratedInput objects.
    """

    def __init__(self, script_path: pathlib.Path):
        super().__init__()
        self.script_path = script_path

    @lark.v_args(inline=False, meta=True)
    def start(self, meta: lark.tree.Meta, statements) -> List[ScriptGeneratedInput]:
        """Return list of ScriptGeneratedInput objects, filtering out None values and flattening nested lists."""
        result = []
        for stmt in statements:
            if stmt is None:
                continue
            elif isinstance(stmt, list):
                # Testgroups return lists, so we need to flatten them
                result.extend(stmt)
            else:
                result.append(stmt)
        return result

    def comment(self, meta: lark.tree.Meta, token: lark.Token) -> None:
        """Comments are ignored and not returned."""
        return None

    def generator_call(
        self, meta: lark.tree.Meta, rest_of_line: str
    ) -> ScriptGeneratedInput:
        """Create ScriptGeneratedInput from a generator call."""
        rest_of_line = rest_of_line.strip()
        name, *arglist = rest_of_line.split(None, 1)
        args = ' '.join(arglist)
        if not args:
            args = None
        return ScriptGeneratedInput(
            generator_call=GeneratorCall(name=name, args=args),
            generator_script=GeneratorScriptEntry(
                path=self.script_path,
                line=meta.line,
            ),
        )

    def generator_name(self, meta: lark.tree.Meta, token: lark.Token) -> str:
        """Return the generator name as string."""
        return str(token)

    def args(self, meta: lark.tree.Meta, rest_of_line: lark.Token) -> str:
        """Return the args string, stripped of leading/trailing whitespace."""
        return str(rest_of_line).strip()

    def copy_test(
        self, meta: lark.tree.Meta, keyword: lark.Token, rest_of_line: str
    ) -> ScriptGeneratedInput:
        """Create ScriptGeneratedInput from a @copy directive."""
        if not rest_of_line.strip():
            raise ValueError('@copy directive requires a filepath')
        return ScriptGeneratedInput(
            copied_from=Testcase(inputPath=pathlib.Path(rest_of_line.strip())),
            generator_script=GeneratorScriptEntry(
                path=self.script_path,
                line=meta.line,
            ),
        )

    @lark.v_args(inline=False, meta=True)
    def inline_input(
        self, meta: lark.tree.Meta, children: List
    ) -> ScriptGeneratedInput:
        """Create ScriptGeneratedInput from an @input directive."""
        # children[0] is the INPUT_KEYWORD token
        # children[1] is either a string or input_lines content
        content = children[1] if len(children) > 1 else ''

        # For brace block syntax (input_lines), strip and ensure trailing newline
        # For string syntax, keep as-is
        # We can distinguish by checking if it's multiline or already processed
        # String syntax (both regular and triple-quoted) don't need modification
        # Brace blocks come from input_lines which already has newlines added per line

        return ScriptGeneratedInput(
            content=content,
            generator_script=GeneratorScriptEntry(
                path=self.script_path,
                line=meta.line,
            ),
        )

    def string(self, meta: lark.tree.Meta, token: lark.Token) -> str:
        """Parse string literal and return its content."""
        raw_string = str(token)

        # Check if it's a triple-quoted string
        if raw_string.startswith('"""') and raw_string.endswith('"""'):
            # Remove triple quotes and return content as-is
            return whitespace.normalize_trailing_lines_from_text(raw_string[3:-3])
        else:
            # Regular string (single or double quoted) - use ast.literal_eval to handle escapes
            return ast.literal_eval(raw_string)

    @lark.v_args(inline=False, meta=True)
    def input_lines(self, meta: lark.tree.Meta, children: List) -> str:
        """Collect all input lines and join them."""
        if not children or children[0] is None:
            return ''
        # children[0] is the BLOCK_CONTENT token
        content = children[0]
        return whitespace.normalize_lines_from_text(str(content))

    @lark.v_args(inline=False, meta=True)
    def testgroup(
        self, meta: lark.tree.Meta, children: List
    ) -> List[ScriptGeneratedInput]:
        """
        Process testgroup and return flattened list of ScriptGeneratedInput objects.

        Children are: TESTGROUP_KEYWORD, GROUP_NAME, statements...
        All statements get the group name assigned to them.
        """
        if len(children) < 2:
            return []

        # First child is TESTGROUP_KEYWORD (@testgroup), second is the group name
        group_name = str(children[1])

        # Rest are statements (can include nested testgroups)
        statements = children[2:]

        # Flatten and assign group name
        result = []
        for stmt in statements:
            if stmt is None:
                continue
            elif isinstance(stmt, list):
                # Nested testgroup returns a list
                for nested_stmt in stmt:
                    if nested_stmt.group is None:
                        nested_stmt.group = group_name
                    result.append(nested_stmt)
            elif isinstance(stmt, ScriptGeneratedInput):
                # Direct statement
                if stmt.group is None:
                    stmt.group = group_name
                result.append(stmt)

        return result


def parse(script: str) -> lark.ParseTree:
    """Parse a test plan script and return the parse tree."""
    return LARK_PARSER.parse(script)


def parse_and_transform(
    script: str, script_path: pathlib.Path
) -> List[ScriptGeneratedInput]:
    """Parse a test plan script and transform it into a list of ScriptGeneratedInput objects."""
    tree = parse(script)
    transformer = TestPlanTransformer(script_path)
    res = transformer.transform(tree)
    return res


if __name__ == '__main__':
    # Example usage
    test_script = """
// This is a comment
gens/generator --MAX_N=100 --MIN_N=30 abcdef

@copy test/in/disk.in

@input '123\\n456\\n789\\n'

@input "test\\ndata"

@input \"\"\"
123
456
789
\"\"\"

@input {
multiline
brace block
syntax
}

@testgroup my-group {
    // Comment inside group
    gens/generator2 --X=5
    gens/generator3 --Y=10 --Z=20 some more args
    @input 'inline in group'
}

@testgroup group_2 {
    gens/generator4 --A=1
}
"""

    # Parse and show tree
    tree = parse(test_script)
    print('Parse tree:')
    print(tree.pretty())
    print()

    # Transform to ScriptGeneratedInput objects
    script_path = pathlib.Path('test_script.txt')
    results = parse_and_transform(test_script, script_path)

    print(f'Generated {len(results)} test inputs:')
    for i, result in enumerate(results, 1):
        line_info = (
            f'Line {result.generator_script.line}'
            if result.generator_script
            else 'Unknown line'
        )
        print(f'\n{i}. {line_info}:')
        print(f'   Group: {result.group}')
        if result.generator_call:
            print(f'   Generator: {result.generator_call.name}')
            print(f'   Args: {result.generator_call.args}')
        if result.copied_from:
            print(f'   Copied from: {result.copied_from.inputPath}')
        if result.content:
            print(f'   Content: {repr(result.content)}')



"""Pytest configuration for E2E tests."""

import pathlib
from typing import Iterator

import pytest

from rbx.box.testing import testing_package


def pytest_configure(config):
    """Configure custom markers."""
    config.addinivalue_line(
        'markers', 'e2e: mark test as end-to-end test (deselect with \'-m "not e2e"\')'
    )
    config.addinivalue_line(
        'markers', 'slow: mark test as slow (deselect with \'-m "not slow"\')'
    )
    config.addinivalue_line('markers', 'docker: mark test as requiring docker')


@pytest.fixture(autouse=True)
def skip_if_no_docker(request):
    """Skip tests marked with 'docker' if docker is not available."""
    if request.node.get_closest_marker('docker'):
        import subprocess

        try:
            subprocess.run(
                ['docker', '--version'],
                check=True,
                capture_output=True,
                timeout=5,
            )
        except (
            subprocess.CalledProcessError,
            FileNotFoundError,
            subprocess.TimeoutExpired,
        ):
            pytest.skip('Docker not available')


@pytest.fixture
def docker_cleanup():
    """Ensure docker containers are cleaned up even on test failure."""
    import atexit
    import subprocess

    containers_to_cleanup = []

    def add_container(container_id):
        containers_to_cleanup.append(container_id)

    def cleanup():
        for container_id in containers_to_cleanup:
            try:
                subprocess.run(
                    ['docker', 'rm', '-f', container_id],
                    capture_output=True,
                )
            except Exception:
                pass

    atexit.register(cleanup)

    yield add_container

    cleanup()


@pytest.fixture
def preset_testing_pkg_from_resources(
    request,
    pkg_from_resources: pathlib.Path,
) -> Iterator[testing_package.TestingPackage]:
    marker = request.node.get_closest_marker('preset_path')
    if marker is None:
        raise ValueError('preset_path marker not found')
    preset_path = pkg_from_resources / marker.args[0]
    with testing_package.TestingPackage(preset_path) as pkg:
        yield pkg



from typing import List, Optional

from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import Label, SelectionList
from textual.widgets.selection_list import Selection


class TabSelectorModal(ModalScreen[Optional[List[int]]]):
    BINDINGS = [
        ('escape', 'cancel', 'Cancel'),
        ('c', 'confirm', 'Confirm'),
        ('a', 'select_all', 'Select all'),
        ('n', 'deselect_all', 'Deselect all'),
    ]

    def __init__(self, tab_names: List[str]) -> None:
        super().__init__()
        self._tab_names = tab_names

    def compose(self) -> ComposeResult:
        with Container(id='tab-selector-dialog'):
            selection_list = SelectionList[int](
                *[
                    Selection(name, index, False)
                    for index, name in enumerate(self._tab_names)
                ],
                id='tab-selector-list',
            )
            selection_list.border_title = 'Select tabs'
            yield selection_list
            yield Label(
                '[b]c[/b] confirm  [b]esc[/b] cancel  '
                '[b]a[/b] select all  [b]n[/b] deselect all',
                id='tab-selector-hints',
            )

    def action_cancel(self) -> None:
        self.dismiss(None)

    def action_confirm(self) -> None:
        selected = list(self.query_one('#tab-selector-list', SelectionList).selected)
        self.dismiss(selected)

    def action_select_all(self) -> None:
        self.query_one('#tab-selector-list', SelectionList).select_all()

    def action_deselect_all(self) -> None:
        self.query_one('#tab-selector-list', SelectionList).deselect_all()



import pathlib
from unittest import mock

import pytest
import typer

from rbx.box import code, package
from rbx.box.schema import CodeItem
from rbx.box.testing import testing_package
from rbx.grading import steps
from rbx.grading.steps import GradingFileInput


class TestCompileItem:
    """Test suite for compile_item function."""

    @pytest.fixture(autouse=True)
    def mock_steps_with_caching(self, testing_pkg: testing_package.TestingPackage):
        """Mock steps_with_caching.compile to avoid heavy operations."""

        def mock_compile_side_effect(
            commands, params, artifacts, sandbox, dependency_cache
        ):
            # Simulate setting digest values for output artifacts
            for output in artifacts.outputs:
                if output.digest is not None:
                    cacher = package.get_file_cacher()
                    # Add the file to the actual file cacher to avoid KeyError
                    # Use the digest returned by put_file_content
                    actual_digest = cacher.put_file_content(b'mock file content')
                    output.digest.value = actual_digest
            return True

        with mock.patch('rbx.box.code.steps_with_caching.compile') as mock_compile:
            mock_compile.side_effect = mock_compile_side_effect
            yield mock_compile

    @pytest.fixture(autouse=True)
    def mock_precompile_header(self):
        """Mock _precompile_header to avoid heavy operations."""
        with mock.patch('rbx.box.code._precompile_header') as mock_precompile:
            mock_precompile.return_value = GradingFileInput(
                src=pathlib.Path('test.h.gch'),
                dest=pathlib.Path('test.h.gch'),
                hash=False,
            )
            yield mock_precompile

    def test_compile_cpp_commands_content(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that C++ compilation commands are properly constructed."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        commands = call_args[0][0]

        assert isinstance(commands, list)
        assert len(commands) == 1
        # Should be the exact C++ compilation command with all modifications
        expected_cmd = 'g++ -std=c++20 -O2 -o executable solution.cpp -fdiagnostics-color=always -Wall -Wshadow -Wno-unused-result -Wno-sign-compare -Wno-char-subscripts -I.'
        assert commands[0] == expected_cmd

    def test_compile_java_commands_content(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that Java compilation commands are properly constructed."""
        java_file = testing_pkg.add_file(
            'Solution.java', src='compile_test/simple.java'
        )
        code_item = CodeItem(path=java_file, language='java')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        commands = call_args[0][0]

        assert isinstance(commands, list)
        assert len(commands) == 2
        # Should be the exact Java compilation commands from default.rbx.yml
        # Notice how the source file gets renamed to Simple.java,
        # which is the name of the Java class contained in Solution.java.
        assert commands[0] == 'javac -Xlint -encoding UTF-8 Simple.java'
        assert commands[1] == 'jar cvf Main.jar @glob:*.class'

    def test_compile_python_no_compilation(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that Python files bypass compilation."""
        py_file = testing_pkg.add_file('solution.py', src='compile_test/simple.py')
        code_item = CodeItem(path=py_file, language='py')

        result = code.compile_item(code_item)

        assert isinstance(result, str)
        assert len(result) > 0
        # Python files don't need compilation
        mock_steps_with_caching.assert_not_called()

    def test_compile_sanitizer_flags_added(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that sanitizer flags are properly added to commands."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item, sanitized=code.SanitizationLevel.FORCE)

        call_args = mock_steps_with_caching.call_args
        commands = call_args[0][0]

        # Should contain sanitizer flags
        assert any('-fsanitize=address,undefined' in cmd for cmd in commands)
        assert any('-fno-omit-frame-pointer' in cmd for cmd in commands)
        assert any('-g' in cmd for cmd in commands)

    def test_compile_warning_flags_added(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that warning flags are properly added to commands."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item, force_warnings=True)

        call_args = mock_steps_with_caching.call_args
        commands = call_args[0][0]

        # Should contain warning flags
        assert any('-Wall' in cmd for cmd in commands)
        assert any('-Wshadow' in cmd for cmd in commands)

    def test_compile_combined_flags(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that sanitizer and warning flags work together."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(
            code_item, sanitized=code.SanitizationLevel.FORCE, force_warnings=True
        )

        call_args = mock_steps_with_caching.call_args
        commands = call_args[0][0]

        # Should contain both sanitizer and warning flags
        assert any('-fsanitize=address,undefined' in cmd for cmd in commands)
        assert any('-Wall' in cmd for cmd in commands)

    def test_compile_artifacts_input_files(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that compilation artifacts contain correct input files."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']

        # Should have input files
        assert len(artifacts.inputs) > 0

        # Should include the main source file (mapped to compilable.cpp)
        main_source_input = next(
            (inp for inp in artifacts.inputs if inp.dest.name == 'solution.cpp'), None
        )
        assert main_source_input is not None
        assert main_source_input.src == cpp_file

        # Should include rbx.h
        rbx_header_input = next(
            (inp for inp in artifacts.inputs if inp.dest.name == 'rbx.h'), None
        )
        assert rbx_header_input is not None

    def test_compile_artifacts_output_files(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that compilation artifacts contain correct output files."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']

        # Should have exactly one output file (the executable)
        assert len(artifacts.outputs) == 1
        output = artifacts.outputs[0]
        assert output.executable is True
        assert output.digest is not None

    def test_compile_non_passthrough_uses_executable_output_src(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """When passthrough is not set, output src should be executable."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']
        output = artifacts.outputs[0]

        assert output.src.name == 'executable'

    def test_compile_passthrough_uses_compilable_output_src(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """When passthrough is set, output src should be compilable (source file)."""
        from rbx.box import environment as env

        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        # Force passthrough in compilation config without mutating cached object
        with mock.patch('rbx.box.code.get_compilation_config') as mock_get_cfg:

            def _side_effect(language, solution=False):
                orig = env.get_compilation_config(language, solution)
                return env.BaseCompilationConfig(
                    commands=orig.commands,
                    sandbox=orig.sandbox,
                    passthrough=True,
                )

            mock_get_cfg.side_effect = _side_effect

            code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']
        output = artifacts.outputs[0]

        # Should point to the compilable (mapped source) instead of executable
        assert output.src.name == 'solution.cpp'

    def test_compile_artifacts_with_testlib(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that testlib.h is included in artifacts when available."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']

        # Should include testlib.h
        testlib_input = next(
            (inp for inp in artifacts.inputs if inp.dest.name == 'testlib.h'), None
        )
        assert testlib_input is not None

    def test_compile_artifacts_with_jngen(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that jngen.h is included in artifacts when available."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']

        # Should include jngen.h
        jngen_input = next(
            (inp for inp in artifacts.inputs if inp.dest.name == 'jngen.h'), None
        )
        assert jngen_input is not None

    def test_compile_sandbox_params_basic(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test basic sandbox parameters setup."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        sandbox_params = call_args.kwargs['params']

        # Should have sandbox parameters
        assert sandbox_params is not None
        # Should have some basic limits for non-sanitized compilation
        assert sandbox_params.timeout is not None
        assert sandbox_params.address_space is not None

    def test_compile_sandbox_params_sanitized_removes_limits(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that sanitized compilation removes memory and time limits."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item, sanitized=code.SanitizationLevel.FORCE)

        call_args = mock_steps_with_caching.call_args
        sandbox_params = call_args.kwargs['params']

        # Sanitized builds should have no memory/time limits
        assert sandbox_params.address_space is None
        assert sandbox_params.timeout is None
        assert sandbox_params.wallclock_timeout is None

    def test_compile_bits_stdcpp_added_for_cpp(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that bits/stdc++.h is added for C++ compilation."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        with mock.patch('rbx.box.code.maybe_get_bits_stdcpp_for_commands') as mock_bits:
            mock_bits.return_value = GradingFileInput(
                src=pathlib.Path('bits/stdc++.h'),
                dest=pathlib.Path('bits/stdc++.h'),
            )

            code.compile_item(code_item)

            call_args = mock_steps_with_caching.call_args
            commands = call_args[0][0]
            artifacts = call_args.kwargs['artifacts']

            # Should call the bits function
            mock_bits.assert_called_once()

            # Should include -I. flag for C++ commands
            assert any('-I.' in cmd for cmd in commands)

            # Should include bits/stdc++.h in artifacts
            bits_input = next(
                (inp for inp in artifacts.inputs if 'stdc++.h' in str(inp.dest)), None
            )
            assert bits_input is not None

    def test_compile_precompilation_enabled_by_default(
        self,
        testing_pkg: testing_package.TestingPackage,
        mock_steps_with_caching,
        mock_precompile_header,
    ):
        """Test that precompilation is enabled by default for C++."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']

        # For simple.cpp, exactly these header files should be present
        expected_header_files = {'testlib.h', 'jngen.h', 'rbx.h', 'stdc++.h'}

        # Find all header files in artifacts
        actual_header_files = {
            inp.dest.name
            for inp in artifacts.inputs
            if inp.dest.suffix in ['.h', '.hpp']
        }

        # Assert exact set of header files
        assert actual_header_files == expected_header_files

        # Only testlib.h, jngen.h, and rbx.h should be processed with warning pragmas
        # stdc++.h is a system header that doesn't get processed
        processed_header_files = {'testlib.h', 'jngen.h', 'rbx.h'}

        for inp in artifacts.inputs:
            if inp.dest.suffix in ['.h', '.hpp']:
                processed_content = inp.src.read_text()
                if inp.dest.name in processed_header_files:
                    # These should have warning pragmas
                    assert '#pragma GCC diagnostic push' in processed_content
                    assert '#pragma GCC diagnostic pop' in processed_content
                    assert '#pragma GCC diagnostic ignored "-Wall"' in processed_content
                    assert (
                        '#pragma GCC diagnostic ignored "-Wshadow"' in processed_content
                    )

    def test_compile_precompilation_disabled(
        self,
        testing_pkg: testing_package.TestingPackage,
        mock_steps_with_caching,
        mock_precompile_header,
    ):
        """Test that precompilation can be disabled."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        code.compile_item(code_item, precompile=False)

        # Should not call precompile_header
        mock_precompile_header.assert_not_called()

    def test_compile_warning_pragmas_processed(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that warning pragmas are added to header files during compilation."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')

        # Create a custom header file that will be processed
        custom_header = testing_pkg.add_file('custom.h')
        custom_header.write_text("""#ifndef CUSTOM_H
#define CUSTOM_H

int custom_function();

#endif""")

        code_item = CodeItem(
            path=cpp_file, language='cpp', compilationFiles=['custom.h']
        )

        code.compile_item(code_item)

        call_args = mock_steps_with_caching.call_args
        artifacts = call_args.kwargs['artifacts']

        # Find the custom header input artifact
        custom_header_input = next(
            (inp for inp in artifacts.inputs if inp.dest.name == 'custom.h'), None
        )
        assert custom_header_input is not None

        # Check that the header file was processed with warning pragmas
        processed_content = custom_header_input.src.read_text()

        # Should contain the exact expected content with warning pragmas wrapped around
        expected_content = """#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wall"
#pragma GCC diagnostic ignored "-Wshadow"
#ifndef CUSTOM_H
#define CUSTOM_H

int custom_function();

#endif
#pragma GCC diagnostic pop
"""

        assert processed_content == expected_content

        # Should be a different file (preprocessed version)
        assert custom_header_input.src != custom_header

    def test_compile_failure_raises_compilation_error(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that compilation failure raises CompilationError."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        with mock.patch('rbx.box.code.steps_with_caching.compile') as mock_compile:
            mock_compile.side_effect = steps.CompilationError()

            with pytest.raises(steps.CompilationError):
                code.compile_item(code_item)

    def test_compile_nonexistent_file_raises_exit(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that compiling nonexistent file raises typer.Exit."""
        nonexistent_file = testing_pkg.path('nonexistent.cpp')
        code_item = CodeItem(path=nonexistent_file, language='cpp')

        with pytest.raises(typer.Exit):
            code.compile_item(code_item)

        mock_steps_with_caching.assert_not_called()

    def test_compile_metadata_set_for_sanitized(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that compilation metadata is set for sanitized builds."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        with mock.patch('rbx.box.code.package.get_file_cacher') as mock_get_cacher:
            mock_cacher = mock.Mock()
            mock_get_cacher.return_value = mock_cacher

            code.compile_item(code_item, sanitized=code.SanitizationLevel.FORCE)

            # Should set metadata indicating sanitized build
            mock_cacher.set_metadata.assert_called_once()
            call_args = mock_cacher.set_metadata.call_args
            assert call_args[0][1] == 'compilation'
            assert call_args[0][2].is_sanitized is True

    def test_compile_metadata_cleared_for_non_sanitized(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that compilation metadata is cleared for non-sanitized builds."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        with mock.patch('rbx.box.code.package.get_file_cacher') as mock_get_cacher:
            mock_cacher = mock.Mock()
            mock_get_cacher.return_value = mock_cacher

            code.compile_item(code_item, sanitized=code.SanitizationLevel.NONE)

            # Should clear metadata for non-sanitized build
            mock_cacher.set_metadata.assert_called_once()
            call_args = mock_cacher.set_metadata.call_args
            assert call_args[0][1] == 'compilation'
            assert call_args[0][2] is None

    def test_compile_verbose_mode_works(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that verbose mode doesn't break compilation."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        # Should not raise any exceptions
        result = code.compile_item(code_item, verbose=True)
        assert isinstance(result, str)
        assert len(result) > 0

    def test_compile_sanitization_level_prefer_vs_force(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test different sanitization levels."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        # Test PREFER level (default)
        code.compile_item(code_item, sanitized=code.SanitizationLevel.PREFER)
        mock_steps_with_caching.assert_called()

        # Test FORCE level
        mock_steps_with_caching.reset_mock()
        code.compile_item(code_item, sanitized=code.SanitizationLevel.FORCE)
        mock_steps_with_caching.assert_called()

        # Test NONE level
        mock_steps_with_caching.reset_mock()
        code.compile_item(code_item, sanitized=code.SanitizationLevel.NONE)
        mock_steps_with_caching.assert_called()

    def test_compile_returns_digest_string(
        self, testing_pkg: testing_package.TestingPackage, mock_steps_with_caching
    ):
        """Test that compile_item returns a valid digest string."""
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        code_item = CodeItem(path=cpp_file, language='cpp')

        result = code.compile_item(code_item)

        # Should return a non-empty string (digest)
        assert isinstance(result, str)
        assert len(result) > 0
        # Digest should be hex-like
        assert all(c in '0123456789abcdef' for c in result.lower())



import pathlib
from typing import List, Optional
from unittest.mock import patch

import pytest

from rbx.box.environment import VerificationLevel
from rbx.box.generators import (
    generate_outputs_for_testcases,
    generate_testcases,
)
from rbx.box.schema import (
    ExpectedOutcome,
    ScoreType,
    Solution,
    Testcase,
)
from rbx.box.solutions import (
    GroupSkeleton,
    SolutionOutcomeStatus,
    SolutionReportSkeleton,
    SolutionSkeleton,
    convert_list_of_solution_evaluations_to_dict,
    get_matching_solutions,
    get_solution_outcome_report,
    run_solutions,
)
from rbx.box.testcase_extractors import extract_generation_testcases_from_groups
from rbx.box.testcase_utils import TestcaseEntry
from rbx.grading.limits import Limits
from rbx.grading.steps import (
    CheckerResult,
    Evaluation,
    Outcome,
    TestcaseIO,
    TestcaseLog,
)


@pytest.mark.test_pkg('problems/box1')
async def test_solutions(pkg_from_testdata: pathlib.Path):
    await generate_testcases()
    entries = [
        entry.group_entry for entry in await extract_generation_testcases_from_groups()
    ]
    await generate_outputs_for_testcases(entries)

    result = run_solutions(verification=VerificationLevel.FULL)
    res = await convert_list_of_solution_evaluations_to_dict(
        result.skeleton, result.items
    )

    # First solution should pass all tests.
    assert all(chk.result.outcome == Outcome.ACCEPTED for chk in res[0]['gen1'])
    # 25 test should be WA for the second solution.
    assert res[1]['gen1'][3].result.outcome == Outcome.WRONG_ANSWER
    # Runtime error for third solution.
    assert all(chk.result.outcome == Outcome.RUNTIME_ERROR for chk in res[2]['gen1'])
    # 1e9 test should be TLE for the fourth solution (soft TLE)
    assert res[3]['gen1'][4].result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    # no TLE outcome should be WA (soft TLE)
    assert res[4]['gen1'][4].result.no_tle_outcome == Outcome.WRONG_ANSWER
    # hard TLE
    assert res[5]['gen1'][4].result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    assert res[5]['gen1'][4].result.no_tle_outcome is None
    # OLE
    assert all(
        chk.result.outcome == Outcome.OUTPUT_LIMIT_EXCEEDED for chk in res[6]['gen1']
    )


@pytest.mark.test_pkg('problems/box1')
async def test_get_solution_outcome_report(pkg_from_testdata: pathlib.Path):
    await generate_testcases()
    entries = [
        entry.group_entry for entry in await extract_generation_testcases_from_groups()
    ]
    await generate_outputs_for_testcases(entries)

    result = run_solutions(verification=VerificationLevel.FULL)
    res = await convert_list_of_solution_evaluations_to_dict(
        result.skeleton, result.items
    )

    # Test AC solution (expected AC, got AC)
    ac_solution = result.skeleton.solutions[0]
    ac_evals = res[0]['gen1']
    ac_report = get_solution_outcome_report(
        ac_solution, result.skeleton, ac_evals, VerificationLevel.FULL
    )
    assert ac_report.status == SolutionOutcomeStatus.OK
    assert ac_report.expectedOutcome == ExpectedOutcome.ACCEPTED
    assert ac_report.gotVerdicts == set()

    # Test WA solution (expected fail, got WA) - should pass since it matches expectation
    wa_solution = result.skeleton.solutions[1]
    wa_evals = res[1]['gen1']
    wa_report = get_solution_outcome_report(
        wa_solution, result.skeleton, wa_evals, VerificationLevel.FULL
    )
    assert wa_report.status == SolutionOutcomeStatus.OK
    assert wa_report.expectedOutcome == ExpectedOutcome.INCORRECT

    # Test RTE solution (expected RTE, got RTE)
    rte_solution = result.skeleton.solutions[2]
    rte_evals = res[2]['gen1']
    rte_report = get_solution_outcome_report(
        rte_solution, result.skeleton, rte_evals, VerificationLevel.FULL
    )
    assert rte_report.status == SolutionOutcomeStatus.OK
    assert rte_report.expectedOutcome == ExpectedOutcome.RUNTIME_ERROR

    # Test TLE solution with double TL warning
    tle_solution = result.skeleton.solutions[3]
    tle_evals = res[3]['gen1']
    tle_report = get_solution_outcome_report(
        tle_solution, result.skeleton, tle_evals, VerificationLevel.FULL
    )
    assert tle_report.status == SolutionOutcomeStatus.OK
    assert tle_report.expectedOutcome == ExpectedOutcome.TIME_LIMIT_EXCEEDED
    # Should have double TL warning for soft TLE
    assert tle_report.runUnderDoubleTl is True or len(tle_report.doubleTlVerdicts) > 0


# Unit tests with custom inputs


@pytest.fixture
def mock_limits():
    """Create mock limits for testing."""
    return Limits(time=1000, memory=256, profile=None, isDoubleTL=False)


@pytest.fixture
def mock_skeleton(tmp_path, mock_limits):
    """Create a minimal skeleton for testing."""

    def _create_skeleton(
        solutions: List[Solution],
        num_entries: int = 5,
    ) -> SolutionReportSkeleton:
        return SolutionReportSkeleton(
            solutions=[
                SolutionSkeleton(**sol.model_dump(), runs_dir=tmp_path / f'run_{i}')
                for i, sol in enumerate(solutions)
            ],
            entries=[TestcaseEntry(group='test', index=i) for i in range(num_entries)],
            groups=[],
            limits={'cpp': mock_limits},
            compiled_solutions={
                str(sol.path): f'digest_{i}' for i, sol in enumerate(solutions)
            },
            verification=VerificationLevel.FULL,
        )

    return _create_skeleton


def make_evaluation(
    outcome: Outcome,
    time_ms: int = 100,
    memory_bytes: int = 1024,
    message: str = '',
    no_tle_outcome: Optional[Outcome] = None,
    sanitizer_warnings: bool = False,
    testcase_index: int = 0,
) -> Evaluation:
    """Helper to create evaluation objects."""
    return Evaluation(
        result=CheckerResult(
            outcome=outcome,
            message=message,
            no_tle_outcome=no_tle_outcome,
            sanitizer_warnings=sanitizer_warnings,
        ),
        log=TestcaseLog(
            time=time_ms / 1000.0,
            memory=memory_bytes,
        ),
        testcase=TestcaseIO(index=testcase_index),
    )


@pytest.fixture
def mock_binary_scoring():
    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.BINARY):
        yield


def test_solution_outcome_report_ac_expects_ac(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test AC solution that expects AC - should pass."""
    solution = Solution(path=tmp_path / 'sol.cpp', outcome=ExpectedOutcome.ACCEPTED)
    skeleton = mock_skeleton([solution])
    evals = [make_evaluation(Outcome.ACCEPTED) for _ in range(5)]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.expectedOutcome == ExpectedOutcome.ACCEPTED
    assert report.gotVerdicts == set()
    assert report.runUnderDoubleTl is False
    assert report.sanitizerWarnings is False


def test_solution_outcome_report_wa_expects_ac(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test WA solution that expects AC - should fail."""
    solution = Solution(path=tmp_path / 'sol.cpp', outcome=ExpectedOutcome.ACCEPTED)
    skeleton = mock_skeleton([solution])
    evals = [
        make_evaluation(Outcome.ACCEPTED),
        make_evaluation(Outcome.ACCEPTED),
        make_evaluation(Outcome.WRONG_ANSWER, message='Expected 5, got 3'),
        make_evaluation(Outcome.ACCEPTED),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.UNEXPECTED_VERDICTS
    assert report.expectedOutcome == ExpectedOutcome.ACCEPTED
    assert Outcome.WRONG_ANSWER in report.gotVerdicts
    assert report.message is not None
    assert report.message[1] == 'Expected 5, got 3'


def test_solution_outcome_report_wa_expects_incorrect(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test WA solution that expects incorrect - should pass."""
    solution = Solution(path=tmp_path / 'wa.cpp', outcome=ExpectedOutcome.INCORRECT)
    skeleton = mock_skeleton([solution])
    evals = [
        make_evaluation(Outcome.ACCEPTED),
        make_evaluation(Outcome.WRONG_ANSWER),
        make_evaluation(Outcome.ACCEPTED),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.expectedOutcome == ExpectedOutcome.INCORRECT


def test_solution_outcome_report_ac_expects_incorrect(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test AC solution that expects incorrect - should fail."""
    solution = Solution(path=tmp_path / 'wa.cpp', outcome=ExpectedOutcome.INCORRECT)
    skeleton = mock_skeleton([solution])
    evals = [make_evaluation(Outcome.ACCEPTED) for _ in range(5)]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.UNEXPECTED_VERDICTS
    assert report.expectedOutcome == ExpectedOutcome.INCORRECT
    assert Outcome.ACCEPTED in report.gotVerdicts


def test_solution_outcome_report_rte_expects_rte(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test RTE solution that expects RTE - should pass."""
    solution = Solution(
        path=tmp_path / 'rte.cpp', outcome=ExpectedOutcome.RUNTIME_ERROR
    )
    skeleton = mock_skeleton([solution])
    evals = [make_evaluation(Outcome.RUNTIME_ERROR) for _ in range(5)]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.expectedOutcome == ExpectedOutcome.RUNTIME_ERROR


def test_solution_outcome_report_tle_with_double_tl(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test TLE solution that runs under double TL - should show warning."""
    solution = Solution(
        path=tmp_path / 'tle.cpp', outcome=ExpectedOutcome.TIME_LIMIT_EXCEEDED
    )
    skeleton = mock_skeleton([solution])
    # Soft TLE (has no_tle_outcome)
    evals = [
        make_evaluation(Outcome.ACCEPTED, time_ms=100),
        make_evaluation(
            Outcome.TIME_LIMIT_EXCEEDED, time_ms=1500, no_tle_outcome=Outcome.ACCEPTED
        ),
        make_evaluation(Outcome.ACCEPTED, time_ms=200),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.expectedOutcome == ExpectedOutcome.TIME_LIMIT_EXCEEDED
    # Should detect it runs under double TL
    assert report.runUnderDoubleTl is True


def test_solution_outcome_report_tle_with_soft_tle_and_wa(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test TLE solution with soft TLE that also has WA in double TL."""
    solution = Solution(
        path=tmp_path / 'tle.cpp', outcome=ExpectedOutcome.TIME_LIMIT_EXCEEDED
    )
    skeleton = mock_skeleton([solution])
    evals = [
        make_evaluation(
            Outcome.TIME_LIMIT_EXCEEDED,
            time_ms=1500,
            no_tle_outcome=Outcome.WRONG_ANSWER,
        ),
        make_evaluation(
            Outcome.TIME_LIMIT_EXCEEDED, time_ms=1600, no_tle_outcome=Outcome.ACCEPTED
        ),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.expectedOutcome == ExpectedOutcome.TIME_LIMIT_EXCEEDED
    # Should show double TL verdicts
    assert len(report.doubleTlVerdicts) > 0
    assert Outcome.WRONG_ANSWER in report.doubleTlVerdicts


def test_solution_outcome_report_sanitizer_warnings(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test solution with sanitizer warnings."""
    solution = Solution(path=tmp_path / 'sol.cpp', outcome=ExpectedOutcome.ACCEPTED)
    skeleton = mock_skeleton([solution])
    evals = [
        make_evaluation(Outcome.ACCEPTED, sanitizer_warnings=False),
        make_evaluation(Outcome.ACCEPTED, sanitizer_warnings=True),
        make_evaluation(Outcome.ACCEPTED, sanitizer_warnings=False),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.sanitizerWarnings is True


def test_solution_outcome_report_subset_mode(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test subset mode shows all verdicts."""
    solution = Solution(path=tmp_path / 'sol.cpp', outcome=ExpectedOutcome.ACCEPTED)
    skeleton = mock_skeleton([solution])
    evals = [
        make_evaluation(Outcome.ACCEPTED),
        make_evaluation(Outcome.ACCEPTED),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL, subset=True
    )

    assert report.status == SolutionOutcomeStatus.OK
    # In subset mode, should show got verdicts even when passing
    assert Outcome.ACCEPTED in report.gotVerdicts
    assert report.expectedOutcome == ExpectedOutcome.ACCEPTED


def test_solution_outcome_report_mixed_outcomes(
    tmp_path, mock_skeleton, mock_binary_scoring
):
    """Test solution with multiple different outcomes."""
    solution = Solution(path=tmp_path / 'sol.cpp', outcome=ExpectedOutcome.ACCEPTED)
    skeleton = mock_skeleton([solution])
    evals = [
        make_evaluation(Outcome.ACCEPTED),
        make_evaluation(Outcome.WRONG_ANSWER),
        make_evaluation(Outcome.RUNTIME_ERROR),
        make_evaluation(Outcome.TIME_LIMIT_EXCEEDED),
    ]

    report = get_solution_outcome_report(
        solution, skeleton, evals, VerificationLevel.FULL
    )

    assert report.status == SolutionOutcomeStatus.UNEXPECTED_VERDICTS
    # Should report the unmatched verdicts
    assert Outcome.WRONG_ANSWER in report.gotVerdicts
    assert Outcome.RUNTIME_ERROR in report.gotVerdicts
    assert Outcome.TIME_LIMIT_EXCEEDED in report.gotVerdicts


def test_get_matching_solutions(tmp_path):
    """Test get_matching_solutions with various filters."""
    # Create mock solutions
    s1 = Solution(
        path=tmp_path / 's1.cpp',
        outcome=ExpectedOutcome.ACCEPTED,
        tags=['implementation', 'easy'],
    )
    s2 = Solution(
        path=tmp_path / 's2.cpp',
        outcome=ExpectedOutcome.TIME_LIMIT_EXCEEDED,
        tags=['brute-force', 'slow'],
    )
    s3 = Solution(
        path=tmp_path / 's3.cpp',
        outcome=ExpectedOutcome.WRONG_ANSWER,
        tags=['implementation', 'buggy'],
    )
    s4 = Solution(
        path=tmp_path / 's4.cpp',
        outcome=ExpectedOutcome.ACCEPTED,
        tags=[],
    )

    with patch(
        'rbx.box.solutions.package.get_solutions', return_value=[s1, s2, s3, s4]
    ):
        # Test no filters
        assert len(get_matching_solutions()) == 4

        # Test filter by expected_outcome
        assert get_matching_solutions(expected_outcome=ExpectedOutcome.ACCEPTED) == [
            s1,
            s4,
        ]
        assert get_matching_solutions(
            expected_outcome=ExpectedOutcome.TIME_LIMIT_EXCEEDED
        ) == [s2]

        # Test filter by tags
        assert get_matching_solutions(tags=['implementation']) == [s1, s3]
        assert get_matching_solutions(tags=['easy']) == [s1]
        assert get_matching_solutions(tags=['brute-force']) == [s2]

        # Test filter by multiple tags (subset check)
        # s1 has implementation and easy.
        assert get_matching_solutions(tags=['implementation', 'easy']) == [s1]
        # order shouldn't matter
        assert get_matching_solutions(tags=['easy', 'implementation']) == [s1]

        # Test non-matching tags
        assert get_matching_solutions(tags=['nonexistent']) == []
        # s1 has implementation but not slow
        assert get_matching_solutions(tags=['implementation', 'slow']) == []

        # Test filter by both outcome and tags
        assert get_matching_solutions(
            expected_outcome=ExpectedOutcome.ACCEPTED, tags=['implementation']
        ) == [s1]

        # s3 is WA and matches implementation
        assert get_matching_solutions(
            expected_outcome=ExpectedOutcome.WRONG_ANSWER, tags=['implementation']
        ) == [s3]

        # s4 is AC but empty tags, shouldn't match if we ask for implementation
        assert get_matching_solutions(
            expected_outcome=ExpectedOutcome.ACCEPTED, tags=['implementation']
        ) == [s1]


def test_solution_outcome_report_points_scoring(tmp_path, mock_limits, mock_skeleton):
    """Test solution reporting with POINTS scoring."""
    # Setup solution with expected score range
    solution = Solution(
        path=tmp_path / 'sol.cpp',
        outcome=ExpectedOutcome.ACCEPTED,
        score=100,  # Expects exactly 100 points
    )

    # Create groups with scores
    g1 = GroupSkeleton(
        name='g1',
        score=30,
        deps=[],
        testcases=[Testcase(inputPath=tmp_path / 'g1_1.in')],
    )
    g2 = GroupSkeleton(
        name='g2',
        score=70,
        deps=[],
        testcases=[Testcase(inputPath=tmp_path / 'g2_1.in')],
    )

    skeleton = SolutionReportSkeleton(
        solutions=[
            SolutionSkeleton(**solution.model_dump(), runs_dir=tmp_path / 'run')
        ],
        entries=[
            TestcaseEntry(group='g1', index=0),
            TestcaseEntry(group='g2', index=0),
        ],
        groups=[g1, g2],
        limits={'cpp': mock_limits},
        compiled_solutions={str(solution.path): 'digest'},
        verification=VerificationLevel.FULL,
    )

    # 1. Test perfect score (30 + 70 = 100)
    evals_perfect = [
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g1
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g2
    ]

    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution, skeleton, evals_perfect, VerificationLevel.FULL
        )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.gotScore == 100
    assert report.maxScore == 100

    # 2. Test partial score (30 + 0 = 30) - Expected 100, got 30 -> UNEXPECTED_SCORE
    evals_partial = [
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g1
        make_evaluation(Outcome.WRONG_ANSWER, testcase_index=0),  # g2
    ]

    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution, skeleton, evals_partial, VerificationLevel.FULL
        )

    assert report.status == SolutionOutcomeStatus.UNEXPECTED_SCORE
    assert report.gotScore == 30

    # 3. Test unexpected score range
    # Solution expects 0..50
    solution_range = Solution(
        path=tmp_path / 'range.cpp',
        outcome=ExpectedOutcome.ANY,
        score=(0, 50),
    )

    # Got 100 (Unlikely for a solution expecting low score, but logic should hold)
    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution_range, skeleton, evals_perfect, VerificationLevel.FULL
        )

    assert report.status == SolutionOutcomeStatus.UNEXPECTED_SCORE
    assert report.gotScore == 100

    # Got 30 (Into range)
    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution_range, skeleton, evals_partial, VerificationLevel.FULL
        )

    assert report.status == SolutionOutcomeStatus.OK
    assert report.gotScore == 30


def test_solution_outcome_report_points_scoring_with_dependencies(
    tmp_path, mock_limits, mock_skeleton
):
    """Test solution reporting with POINTS scoring and dependencies."""
    solution = Solution(
        path=tmp_path / 'sol.cpp',
        outcome=ExpectedOutcome.ACCEPTED,
        score=100,
    )

    # Create groups with dependencies
    # g1 (30)
    # g2 (30) -> deps: g1
    # g3 (40) -> deps: g2
    g1 = GroupSkeleton(
        name='g1',
        score=30,
        deps=[],
        testcases=[Testcase(inputPath=tmp_path / 'g1_1.in')],
    )
    g2 = GroupSkeleton(
        name='g2',
        score=30,
        deps=['g1'],
        testcases=[Testcase(inputPath=tmp_path / 'g2_1.in')],
    )
    g3 = GroupSkeleton(
        name='g3',
        score=40,
        deps=['g2'],
        testcases=[Testcase(inputPath=tmp_path / 'g3_1.in')],
    )

    skeleton = SolutionReportSkeleton(
        solutions=[
            SolutionSkeleton(**solution.model_dump(), runs_dir=tmp_path / 'run')
        ],
        entries=[
            TestcaseEntry(group='g1', index=0),
            TestcaseEntry(group='g2', index=0),
            TestcaseEntry(group='g3', index=0),
        ],
        groups=[g1, g2, g3],
        limits={'cpp': mock_limits},
        compiled_solutions={str(solution.path): 'digest'},
        verification=VerificationLevel.FULL,
    )

    # 1. All pass -> Score 100
    evals_all_pass = [
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g1
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g2
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g3
    ]

    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution, skeleton, evals_all_pass, VerificationLevel.FULL
        )
    assert report.gotScore == 100

    # 2. g1 fails -> Score 0 (g2 and g3 check deps and fail efficiently or just don't count)
    evals_g1_fail = [
        make_evaluation(Outcome.WRONG_ANSWER, testcase_index=0),  # g1
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g2
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g3
    ]

    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution, skeleton, evals_g1_fail, VerificationLevel.FULL
        )
    assert report.gotScore == 0

    # 3. g1 passes, g2 fails -> Score 30 (g3 blocked)
    evals_g2_fail = [
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g1
        make_evaluation(Outcome.WRONG_ANSWER, testcase_index=0),  # g2
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g3
    ]

    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution, skeleton, evals_g2_fail, VerificationLevel.FULL
        )
    assert report.gotScore == 30

    # 4. g1 passes, g2 passes, g3 fails -> Score 60
    evals_g3_fail = [
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g1
        make_evaluation(Outcome.ACCEPTED, testcase_index=0),  # g2
        make_evaluation(Outcome.WRONG_ANSWER, testcase_index=0),  # g3
    ]

    with patch('rbx.box.solutions.package.get_scoring', return_value=ScoreType.POINTS):
        report = get_solution_outcome_report(
            solution, skeleton, evals_g3_fail, VerificationLevel.FULL
        )
    assert report.gotScore == 60



import pytest
import typer

from rbx.box.code import maybe_rename_java_class
from rbx.box.environment import FileMapping


class TestMaybeRenameJavaClass:
    def test_basic_class_renaming(self, testing_pkg):
        """Test basic class renaming from Solution to Main"""
        java_file = testing_pkg.add_file('Solution.java')
        java_file.write_text("""public class Solution {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}""")

        file_mapping = FileMapping(compilable='Main.java', executable='Main')
        result_path = maybe_rename_java_class(java_file, file_mapping)

        expected_content = """public class Main {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}"""

        assert result_path.read_text() == expected_content
        assert result_path != java_file  # Should create new file

    def test_no_change_needed(self, testing_pkg):
        """Test when class is already named correctly"""
        java_file = testing_pkg.add_file('Main.java')
        original_content = """public class Main {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}"""
        java_file.write_text(original_content)

        file_mapping = FileMapping(compilable='Main.java', executable='Main')
        result_path = maybe_rename_java_class(java_file, file_mapping)

        assert result_path.read_text() == original_content
        assert result_path == java_file  # Should return original file

    def test_whitespace_variations(self, testing_pkg):
        """Test handling of various whitespace patterns"""
        java_file = testing_pkg.add_file('MyClass.java')
        java_file.write_text("""public   class    MyClass   {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}""")

        file_mapping = FileMapping(compilable='Main.java', executable='Main')
        result_path = maybe_rename_java_class(java_file, file_mapping)

        expected_content = """public class Main   {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}"""

        assert result_path.read_text() == expected_content

    def test_complex_class_name(self, testing_pkg):
        """Test renaming with complex class names containing underscores, numbers, and dollar signs"""
        java_file = testing_pkg.add_file('Complex_Class_123$.java')
        java_file.write_text("""public class Complex_Class_123$ {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}""")

        file_mapping = FileMapping(compilable='Main.java', executable='Main')
        result_path = maybe_rename_java_class(java_file, file_mapping)

        expected_content = """public class Main {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}"""

        assert result_path.read_text() == expected_content

    def test_no_public_class_raises_error(self, testing_pkg):
        """Test that missing public class raises typer.Exit"""
        java_file = testing_pkg.add_file('NoPublic.java')
        java_file.write_text("""class NoPublic {
    public static void main(String[] args) {
        System.out.println("Hello World");
    }
}""")

        file_mapping = FileMapping(compilable='Main.java', executable='Main')

        with pytest.raises(typer.Exit):
            maybe_rename_java_class(java_file, file_mapping)

    def test_multiple_classes_only_renames_public(self, testing_pkg):
        """Test that only the public class gets renamed when multiple classes exist"""
        java_file = testing_pkg.add_file('MultiClass.java')
        java_file.write_text("""class Helper {
    public void help() {
        System.out.println("Helping");
    }
}

public class MultiClass {
    public static void main(String[] args) {
        Helper h = new Helper();
        h.help();
    }
}

class AnotherHelper {
    public void assist() {
        System.out.println("Assisting");
    }
}""")

        file_mapping = FileMapping(compilable='Main.java', executable='Main')
        result_path = maybe_rename_java_class(java_file, file_mapping)

        expected_content = """class Helper {
    public void help() {
        System.out.println("Helping");
    }
}

public class Main {
    public static void main(String[] args) {
        Helper h = new Helper();
        h.help();
    }
}

class AnotherHelper {
    public void assist() {
        System.out.println("Assisting");
    }
}"""

        assert result_path.read_text() == expected_content

    def test_preserves_comments_and_imports(self, testing_pkg):
        """Test that comments and imports are preserved during renaming"""
        java_file = testing_pkg.add_file('Solution.java')
        java_file.write_text("""// This is a comment
import java.util.*;
import java.io.*;

/* Multi-line comment
   about the solution */
public class Solution {
    // Another comment
    public static void main(String[] args) {
        System.out.println("Hello World"); // Inline comment
    }
}""")

        file_mapping = FileMapping(compilable='Main.java', executable='Main')
        result_path = maybe_rename_java_class(java_file, file_mapping)

        expected_content = """// This is a comment
import java.util.*;
import java.io.*;

/* Multi-line comment
   about the solution */
public class Main {
    // Another comment
    public static void main(String[] args) {
        System.out.println("Hello World"); // Inline comment
    }
}"""

        assert result_path.read_text() == expected_content

    def test_non_java_files_unaffected(self, testing_pkg):
        """Test that non-Java files are returned unchanged"""
        cpp_file = testing_pkg.add_file('solution.cpp')
        original_content = """#include <iostream>
using namespace std;

int main() {
    cout << "Hello World" << endl;
    return 0;
}"""
        cpp_file.write_text(original_content)

        file_mapping = FileMapping(compilable='solution.cpp', executable='solution')
        result_path = maybe_rename_java_class(cpp_file, file_mapping)

        assert result_path.read_text() == original_content
        assert result_path == cpp_file  # Should return original file unchanged



"""Registry of supported LLM models and their settings for the wizard service."""

from agents.model_settings import ModelSettings

# Map of supported model name to its ModelSettings
SUPPORTED_MODELS: dict[str, ModelSettings] = {
    # OpenAI family
    'gpt-4o': ModelSettings(temperature=0.1),
    'gpt-4.1': ModelSettings(temperature=0.1),
    # Default ROBX choices
    'gpt-5': ModelSettings(),
    'gpt-5-mini': ModelSettings(),
}


def get_model_settings(model_name: str) -> ModelSettings:
    """Return settings for a supported model or raise a ValueError.

    Args:
        model_name: Name of the model to look up.

    Raises:
        ValueError: If the model is not supported.
    """

    try:
        return SUPPORTED_MODELS[model_name]
    except KeyError as exc:
        supported = ', '.join(sorted(SUPPORTED_MODELS.keys()))
        raise ValueError(
            f"Unsupported model '{model_name}'. Supported models: {supported}"
        ) from exc


def is_supported_model(model_name: str) -> bool:
    """Check whether a model is in the supported registry."""

    return model_name in SUPPORTED_MODELS


# Default model that routes will use unless specified otherwise.
DEFAULT_MODEL_NAME = 'gpt-4.1'



import asyncio
import difflib
import json
import tempfile
from pathlib import Path
from typing import Any, Callable, List, Optional

from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.coordinate import Coordinate
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import DataTable, Footer, Header, Input, Select, Static

from rbx.box.schema import ExpectedOutcome
from rbx.box.solutions import (
    get_full_ui_friendly_outcome_markup_verdict,
)
from rbx.box.tooling.boca.scraper import (
    BocaDetailedRun,
    BocaProblem,
    BocaRun,
    BocaScraper,
    get_boca_scraper,
)
from rbx.box.ui.widgets.code_box import CodeBox
from rbx.box.ui.widgets.diff_box import DiffBox
from rbx.config import get_app_path
from rbx.grading.steps import Outcome


def _format_time(minutes: int) -> str:
    hours = minutes // 60
    mins = minutes % 60
    return f'{hours:02d}:{mins:02d}'


class BocaRunsApp(App):
    CSS_PATH = None

    # Compact layout styling for filters, mode indicator and teams panel
    CSS = """
        #left_panel {
            width: 3fr;
        }
        #right_panel {
            width: 3fr;
        }
            #loading_indicator {
                height: 1;
                padding: 0 1;
                margin: 0 1;
            }
        #filters {
            padding: 0 1;
            margin: 0 1;
        }
        #filters_row1 { }
        #filters_row2 { }
        #problem_select { width: 24; }
        #verdict_select { width: 20; }
        #refresh_label { width: 6; }
        #refresh_input { width: 12; }
        #diff_label { width: 4; }
        #diff_input { width: 10; }
            #team_input { width: 1fr; }
        #mode_indicator {
            height: 1;
            padding: 0 1;
            margin: 0 1;
        }
        #runs_table { height: 3fr; }
        """

    mode: reactive[str] = reactive('judged')
    pending_requests: reactive[int] = reactive(0)
    verdict_filter: reactive[Optional[str]] = reactive(None)
    problem_filter: reactive[Optional[str]] = reactive(None)
    team_filter: reactive[Optional[str]] = reactive(None)
    refresh_interval: reactive[int] = reactive(30)
    contest_id: reactive[Optional[str]] = reactive(None)

    def __init__(
        self, scraper: Optional[BocaScraper] = None, contest_id: Optional[str] = None
    ):
        super().__init__()
        self.scraper = scraper or get_boca_scraper()
        self._problems: List[BocaProblem] = []
        self._runs: List[BocaRun] = []
        # Indexes for fast lookup/filtering
        self._runs_by_problem: dict[str, List[BocaRun]] = {}
        self._runs_by_team: dict[str, List[BocaRun]] = {}
        self._runs_by_problem_user: dict[tuple[str, str], List[BocaRun]] = {}
        self._tmp_dir = Path(tempfile.gettempdir()) / 'rbx_boca_viewer'
        self._tmp_dir.mkdir(parents=True, exist_ok=True)
        self._selected_key: Optional[str] = None
        self._highlighted_key: Optional[str] = None
        # Background tasks
        self._refresh_task: Optional[asyncio.Task] = None
        # Persistent cache
        self._cache_base: Optional[Path] = None
        self._runs_dir: Optional[Path] = None
        self._prefetch_inflight: set[str] = set()
        # Small diff detection cache: run_key -> is_small_diff
        self._small_diff_flags: dict[str, bool] = {}
        # Prioritized scraper executor (serializes calls, honors priority)
        self.PRIORITY_USER = 0
        self.PRIORITY_NORMAL = 5
        self.PRIORITY_PREFETCH = 10
        self._scraper_queue: 'asyncio.PriorityQueue[tuple[int, int, Callable[..., Any], tuple[Any, ...], dict[str, Any], asyncio.Future]]' = asyncio.PriorityQueue()  # type: ignore[type-arg]
        self._scraper_seq: int = 0
        self._scraper_worker_task: Optional[asyncio.Task] = None
        # Small diff result cache: run_key -> metadata/results
        # Used to avoid recomputing on every refresh and to drive UI status.
        self._small_diff_cache: dict[str, dict[str, Any]] = {}
        # Contest id provided externally (defaults to 'default' if not provided)
        self.contest_id = str(contest_id).strip() if contest_id else 'default'
        # Track last "team zoom" to allow restoring previous filter
        self._team_zoom_prev_filter: Optional[str] = None
        self._team_zoom_active_team: Optional[str] = None

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        # Global loading indicator (reflects any in-flight network requests)
        yield Static('', id='loading_indicator')
        with Horizontal():
            with Vertical(id='left_panel'):
                yield Static('', id='mode_indicator')
                with Vertical(id='filters'):
                    with Horizontal(id='filters_row1'):
                        yield Select([], prompt='Problem', id='problem_select')
                        yield Select([], prompt='Verdict', id='verdict_select')
                    with Horizontal(id='filters_row2'):
                        yield Static('Ref:', id='refresh_label')
                        yield Input(
                            str(self.refresh_interval),
                            id='refresh_input',
                            placeholder='30',
                        )
                        yield Static('Δ:', id='diff_label')
                        yield Input(
                            str(self.small_diff_threshold),
                            id='diff_input',
                            placeholder='5',
                        )
                        yield Input(
                            placeholder='Team filter (substring, case-insensitive)',
                            id='team_input',
                        )
                table = DataTable(id='runs_table')
                table.add_columns(
                    'Run',
                    'Site',
                    'Problem',
                    'Verdict',
                    'Time',
                    'Status',
                    'Team',
                    'BG',
                )
                table.cursor_type = 'row'
                yield table
                yield Static('', id='left_spacer')
            with Vertical(id='right_panel'):
                # Shows which submission is currently selected
                yield Static('No run selected.', id='selection_info')
                self.code_box = CodeBox()
                yield self.code_box
        yield Footer()

    async def on_mount(self) -> None:
        # Initialize cache paths based on pre-configured contest id
        self._init_cache_paths()
        self.log(f'on_mount: using contest_id={self.contest_id!r}')
        # Run initial login + refresh in background to avoid blocking UI render
        asyncio.create_task(self._initial_load())
        self._update_mode_indicator()
        # Set initial focus to the runs table
        try:
            table = self.query_one('#runs_table', DataTable)
            self.set_focus(table)
            self.log('on_mount: focus set to runs_table')
        except Exception:
            self.log('on_mount: failed to set focus to runs_table')
        self.log('on_mount: completed')
        # Start auto-refresh loop and scraper worker
        self._start_auto_refresh_task()
        self._start_scraper_worker()

    # Contest id is now provided externally (CLI); in-app prompt removed

    async def _initial_load(self) -> None:
        try:
            await self._ensure_login()
            await self._refresh_runs()
        except Exception:
            # Errors are logged by Textual's default handler; ensure indicator decrements
            self.log('_initial_load: encountered exception during startup')

    async def _ensure_login(self) -> None:
        self.log('ensure_login: logging into BOCA via scraper')
        await self._run_scraper(self.scraper.login, priority=self.PRIORITY_USER)
        self.log('ensure_login: login complete')

    def _populate_filters(self) -> None:
        problem_select = self.query_one('#problem_select', Select)
        verdict_select = self.query_one('#verdict_select', Select)

        problem_values = sorted({r.problem_shortname for r in self._runs})
        problem_options = [('All', '__all__')] + [
            (name, name) for name in problem_values
        ]
        # Prefer the stored filter value if present; otherwise use current widget value
        current_problem_value = (
            self.problem_filter
            if self.problem_filter is not None
            else getattr(problem_select, 'value', None)
        )
        problem_select.set_options(problem_options)
        valid_values = {v for _, v in problem_options}
        problem_select.value = (
            current_problem_value
            if current_problem_value in valid_values
            else '__all__'
        )

        verdict_options = [('All', '__all__')] + [
            (eo.name.replace('_', ' ').title(), eo.name) for eo in ExpectedOutcome
        ]
        verdict_select.set_options(verdict_options)
        # Preserve verdict selection similarly
        current_verdict_value = (
            self.verdict_filter
            if self.verdict_filter is not None
            else getattr(verdict_select, 'value', None)
        )
        verdict_valid_values = {v for _, v in verdict_options}
        verdict_select.value = (
            current_verdict_value
            if current_verdict_value in verdict_valid_values
            else '__all__'
        )

    async def _refresh_runs(self) -> None:
        self.log(f'_refresh_runs: fetching runs (mode={self.mode})')
        only_judged = True if self.mode == 'judged' else False
        runs = await self._run_scraper(
            self.scraper.list_runs, only_judged, priority=self.PRIORITY_NORMAL
        )
        if self.mode == 'queue':
            runs = [r for r in runs if (r.outcome is None or r.status != 'judged')]
        elif self.mode == 'judged':
            runs = [r for r in runs if r.outcome is not None]
        else:  # both
            # Keep all returned runs
            pass
        self._runs = runs
        self.log(f'_refresh_runs: fetched {len(self._runs)} runs')
        # rebuild indexes for fast queries
        self._rebuild_run_indexes()
        # refresh problem filter based on current runs
        self._populate_filters()
        self._reload_table()
        self._update_mode_indicator()
        # Prefetch codes for runs not yet cached (non-blocking)
        asyncio.create_task(self._prefetch_missing_runs())
        # Compute small-diff markers in background (non-blocking)
        asyncio.create_task(self._compute_small_diffs())

    def _rebuild_run_indexes(self) -> None:
        by_problem: dict[str, List[BocaRun]] = {}
        by_team: dict[str, List[BocaRun]] = {}
        by_problem_user: dict[tuple[str, str], List[BocaRun]] = {}
        for run in self._runs:
            problem = run.problem_shortname
            team = (run.user or '').strip()
            by_problem.setdefault(problem, []).append(run)
            by_team.setdefault(team, []).append(run)
            by_problem_user.setdefault((problem, team), []).append(run)

        # Sort each list to allow efficient previous-run scans
        def _key_fn(r: BocaRun):
            return (r.time, r.run_number)

        for lst in by_problem.values():
            lst.sort(key=_key_fn)
        for lst in by_team.values():
            lst.sort(key=_key_fn)
        for lst in by_problem_user.values():
            lst.sort(key=_key_fn)
        self._runs_by_problem = by_problem
        self._runs_by_team = by_team
        self._runs_by_problem_user = by_problem_user

    def _passes_filters(self, run: BocaRun) -> bool:
        if self.problem_filter and self.problem_filter != '__all__':
            if run.problem_shortname != self.problem_filter:
                return False
        if self.team_filter:
            team_name = (run.user or '').strip()
            if self.team_filter.lower() not in team_name.lower():
                return False
        if self.mode in ('judged', 'both'):
            if self.verdict_filter and self.verdict_filter != '__all__':
                try:
                    expected = ExpectedOutcome[self.verdict_filter]
                except KeyError:
                    return False
                if run.outcome is None:
                    return False
                if not expected.match(run.outcome):
                    return False
        return True

    def _reload_table(self) -> None:
        table = self.query_one('#runs_table', DataTable)
        # Remember current highlighted key to restore after reload
        prev_highlight = self._highlighted_key
        table.clear()
        rows_added = 0
        first_key: Optional[str] = None
        for run in self._runs:
            if not self._passes_filters(run):
                continue
            if run.outcome is not None:
                verdict = Text.from_markup(
                    get_full_ui_friendly_outcome_markup_verdict(run.outcome)
                )
            else:
                verdict = '—'
            time_s = _format_time(run.time)
            row_key = f'{run.run_number}:{run.site_number}'
            self.log(
                f'_reload_table: add row key={row_key} problem={run.problem_shortname} team={(run.user or "").strip()} verdict={verdict} status={run.status}'
            )
            star = ''
            if run.outcome is not None and run.outcome == Outcome.ACCEPTED:
                flag = self._small_diff_flags.get(row_key)
                if flag:
                    star = '[red]*[/red]'
            table.add_row(
                f'{run.run_number}{star}',
                str(run.site_number),
                run.problem_shortname,
                verdict,
                time_s,
                run.status,
                run.user or '',
                self._bg_status_for_run(run),
                key=row_key,
            )
            rows_added += 1
            if first_key is None:
                first_key = row_key
        self.log(f'_reload_table: added {rows_added} rows')
        # Try to restore previously highlighted row if it still exists
        if prev_highlight is not None:
            try:
                row_idx = table.get_row_index(prev_highlight)
                table.cursor_coordinate = Coordinate(row=row_idx, column=0)
                self._highlighted_key = prev_highlight
            except Exception:
                # Fallback to first row if previous no longer exists
                if first_key is not None:
                    try:
                        row_idx = table.get_row_index(first_key)
                        table.cursor_coordinate = Coordinate(row=row_idx, column=0)
                        self._highlighted_key = first_key
                    except Exception:
                        # As a last resort, just record the key
                        self._highlighted_key = first_key
        else:
            if first_key is not None:
                try:
                    row_idx = table.get_row_index(first_key)
                    table.cursor_coordinate = Coordinate(row=row_idx, column=0)
                except Exception:
                    pass
                self._highlighted_key = first_key

    # --- Auto-refresh management ---
    def _cancel_auto_refresh_task(self) -> None:
        task = self._refresh_task
        if task is not None and not task.done():
            task.cancel()
        self._refresh_task = None

    def _start_auto_refresh_task(self) -> None:
        self._cancel_auto_refresh_task()

        async def _runner() -> None:
            while True:
                try:
                    await asyncio.sleep(max(1, int(self.refresh_interval)))
                    if self.refresh_interval > 0:
                        await self._refresh_runs()
                    else:
                        await asyncio.sleep(1)
                except asyncio.CancelledError:
                    break
                except Exception:
                    # Ignore intermittent failures and keep the loop running
                    pass

        self._refresh_task = asyncio.create_task(_runner())

    def watch_refresh_interval(self, value: int) -> None:  # called on change
        # Restart the loop to apply the new interval immediately
        self._start_auto_refresh_task()

    async def action_toggle_mode(self) -> None:
        if self.mode == 'judged':
            self.mode = 'queue'
        elif self.mode == 'queue':
            self.mode = 'both'
        else:
            self.mode = 'judged'
        self._update_mode_indicator()
        await self._refresh_runs()

    BINDINGS = [
        ('r', 'refresh', 'Refresh'),
        ('m', 'toggle_mode', 'Toggle mode'),
        ('d', 'show_diff', 'Last diff'),
        ('D', 'show_non_ac_diff', 'Last non-AC diff'),
        ('p', 'toggle_problem_filter', 'Toggle problem filter'),
        ('t', 'toggle_team_filter', 'Toggle team filter'),
        ('q', 'quit', 'Quit'),
    ]

    def _extract_key_str(self, key: object) -> Optional[str]:
        if key is None:
            self.log('_extract_key_str: key is None')
            return None
        if isinstance(key, str):
            self.log(f'_extract_key_str: key is str -> {key}')
            return key
        # Extract the underlying string from Textual's RowKey wrapper when necessary
        value = getattr(key, 'value', None)
        if isinstance(value, str):
            self.log(f'_extract_key_str: key has .value -> {value}')
            return value
        try:
            # Fallback to string conversion
            as_str = str(key)
            if ':' in as_str:
                self.log(f'_extract_key_str: key fallback str -> {as_str}')
                return as_str
            self.log(f'_extract_key_str: key fallback str has no colon -> {as_str}')
            return None
        except Exception:
            self.log('_extract_key_str: exception while converting key to str')
            return None

    def _run_from_row_key(self, key: object) -> Optional[BocaRun]:
        key_str = self._extract_key_str(key)
        if key_str is None or ':' not in key_str:
            self.log(f'_run_from_row_key: invalid key_str -> {key_str}')
            return None
        run_number_str, site_number_str = key_str.split(':', 1)
        run = next(
            (
                r
                for r in self._runs
                if str(r.run_number) == run_number_str
                and str(r.site_number) == site_number_str
            ),
            None,
        )
        if run is None:
            self.log(f'_run_from_row_key: no run found for key {key_str}')
        else:
            self.log(
                f'_run_from_row_key: resolved run {run.run_number}-{run.site_number} ({run.problem_shortname})'
            )
        return run

    def _update_selection_info(
        self,
        run: BocaRun,
        detailed: Optional[BocaDetailedRun] = None,
        *,
        loading: bool = False,
    ) -> None:
        info = []
        verdict = run.outcome.name if run.outcome is not None else '—'
        team = (run.user or '').strip()
        info.append(
            f'Run {run.run_number}-{run.site_number} | Problem {run.problem_shortname} | Team {team}'
        )
        info.append(
            f'Verdict {verdict} | Time {_format_time(run.time)} | Status {run.status}'
        )
        if detailed is not None:
            info.append(f'File {detailed.filename.name}')
        if loading:
            info.append('⏳ [yellow]Loading code…[/yellow]')
        self.query_one('#selection_info', Static).update('\n'.join(info))

    def _update_mode_indicator(self) -> None:
        indicator = self.query_one('#mode_indicator', Static)
        label = {
            'judged': 'Judged',
            'queue': 'Queue',
            'both': 'Both',
        }.get(self.mode, self.mode)
        indicator.update(
            f'Contest: {self.contest_id or "?"} | Mode: {label} (press "m" to change)'
        )

    async def action_refresh(self) -> None:
        await self._refresh_runs()

    def action_toggle_problem_filter(self) -> None:
        # Toggle between "All" and highlighted run's problem
        try:
            problem_select = self.query_one('#problem_select', Select)
        except Exception:
            return
        # Zoom out first regardless of highlight state
        if self.problem_filter and self.problem_filter != '__all__':
            self.problem_filter = '__all__'
            problem_select.value = '__all__'
            self._reload_table()
            return
        # Zoom in requires a highlighted row
        key_str = self._highlighted_key
        run = self._run_from_row_key(key_str) if key_str is not None else None
        if run is None:
            try:
                self.notify('Highlight a run first.', severity='error')
            except Exception:
                pass
            return
        self.problem_filter = run.problem_shortname
        problem_select.value = run.problem_shortname
        self._reload_table()

    def action_toggle_team_filter(self) -> None:
        # First press: save current filter and zoom to highlighted team.
        # Second press: restore the saved filter.
        try:
            team_input = self.query_one('#team_input', Input)
        except Exception:
            return
        # Zoom out (restore previous) does not require highlight
        if self._team_zoom_prev_filter is not None:
            prev = self._team_zoom_prev_filter
            self._team_zoom_prev_filter = None
            self._team_zoom_active_team = None
            team_input.value = prev
            self.team_filter = prev if prev else None
            self._reload_table()
            return
        # Zoom in requires a highlighted row
        key_str = self._highlighted_key
        run = self._run_from_row_key(key_str) if key_str is not None else None
        if run is None:
            try:
                self.notify('Highlight a run first.', severity='error')
            except Exception:
                pass
            return
        team_name = (run.user or '').strip()
        # Save current filter and zoom into highlighted team
        current_value = (team_input.value or '').strip()
        self._team_zoom_prev_filter = current_value
        self._team_zoom_active_team = team_name
        team_input.value = team_name
        self.team_filter = team_name if team_name else None
        self._reload_table()

    def action_quit(self) -> None:
        self.exit()

    async def action_show_options(self) -> None:
        # Open options without waiting; handle result in on_screen_dismissed
        try:
            self.push_screen(
                _OptionsScreen(self.refresh_interval, self.small_diff_threshold)
            )
        except Exception:
            return

    def on_screen_dismissed(self, event) -> None:  # type: ignore[override]
        # Capture result from options modal
        if isinstance(event.screen, _OptionsScreen):
            result = getattr(event, 'result', None)
            self.log(f'on_screen_dismissed: options result={result!r}')
            if not isinstance(result, dict):
                return
            try:
                new_refresh = int(result.get('refresh_interval', self.refresh_interval))
                if new_refresh > 0:
                    self.refresh_interval = new_refresh
                    self.log(f'on_screen_dismissed: applied refresh={new_refresh}')
            except Exception:
                pass
            try:
                new_delta = int(
                    result.get('small_diff_threshold', self.small_diff_threshold)
                )
                if new_delta > 0:
                    self.small_diff_threshold = new_delta
                    self.log(f'on_screen_dismissed: applied diff={new_delta}')
            except Exception:
                pass
            try:
                self.notify('Options saved', severity='information')
            except Exception:
                pass

    async def action_show_non_ac_diff(self) -> None:
        await self.action_show_diff(last_non_ac=True)

    async def action_show_diff(self, last_non_ac: bool = False) -> None:
        # Use currently highlighted row; selection not required
        key_str = self._highlighted_key
        if key_str is None:
            self.notify('Highlight a run first to diff.', severity='error')
            return
        run = self._run_from_row_key(key_str)
        if run is None:
            self.notify('Could not resolve selected run.', severity='error')
            return

        team = (run.user or '').strip()
        problem = run.problem_shortname
        site_number = run.site_number

        # Find latest previous run according to chosen mode
        if last_non_ac:
            prev = self._find_last_non_ac_before(run)
            if prev is None:
                self.notify(
                    'No previous non-AC run found for this team and problem.',
                    severity='error',
                )
                return
        else:
            # Use indexed runs for (problem, team) and filter by site/time
            indexed = self._runs_by_problem_user.get((problem, team), [])
            candidates = [
                r
                for r in indexed
                if r.site_number == site_number
                and r.outcome is not None
                and (
                    (r.time < run.time)
                    or (r.time == run.time and r.run_number < run.run_number)
                )
            ]
            if not candidates:
                self.notify(
                    'No previous run found for this team and problem.', severity='error'
                )
                return
            prev = max(candidates, key=lambda r: (r.time, r.run_number))

        try:
            path_current, path_prev = await asyncio.gather(
                self._ensure_cached_run(
                    run, is_prefetch=False, priority=self.PRIORITY_USER
                ),
                self._ensure_cached_run(
                    prev, is_prefetch=False, priority=self.PRIORITY_USER
                ),
            )
        except Exception:
            self.notify('Failed to prepare runs for diff.', severity='error')
            return

        if path_current is None or path_prev is None:
            self.notify('Failed to prepare files for diff.', severity='error')
            return

        self.push_screen(_BocaDifferScreen(path_prev, path_current))

    @on(Select.Changed, '#problem_select')
    async def _on_problem_changed(self, event: Select.Changed) -> None:
        value = event.value
        self.problem_filter = None if value is None else str(value)
        self._reload_table()

    @on(Input.Changed, '#refresh_input')
    def _on_refresh_interval_changed(self, event: Input.Changed) -> None:
        raw = (event.value or '').strip()
        try:
            new_value = int(raw)
            if new_value <= 0:
                return
            self.refresh_interval = new_value
        except ValueError:
            # Ignore non-integer inputs
            return

    @on(Select.Changed, '#verdict_select')
    async def _on_verdict_changed(self, event: Select.Changed) -> None:
        value = event.value
        self.verdict_filter = None if value is None else str(value)
        self._reload_table()

    @on(Input.Changed, '#diff_input')
    def _on_small_diff_threshold_changed(self, event: Input.Changed) -> None:
        raw = (event.value or '').strip()
        try:
            new_value = int(raw)
            if new_value <= 0:
                return
            self.small_diff_threshold = new_value
        except ValueError:
            return

    @on(Input.Changed, '#team_input')
    def _on_team_filter_changed(self, event: Input.Changed) -> None:
        raw = (event.value or '').strip()
        self.team_filter = raw if raw else None
        self._reload_table()

    # --- Small diff threshold configuration ---
    small_diff_threshold: reactive[int] = reactive(5)

    def watch_small_diff_threshold(self, value: int) -> None:  # called on change
        # Recompute markers with new threshold
        asyncio.create_task(self._compute_small_diffs())

    @on(DataTable.RowSelected, '#runs_table')
    async def _on_row_selected(self, event: DataTable.RowSelected) -> None:
        self.log(f'RowSelected: event.row_key={event.row_key!r}')
        key_str = self._extract_key_str(event.row_key)
        run = self._run_from_row_key(event.row_key)
        if run is None:
            self.log('RowSelected: run is None (ignoring)')
            return
        # Remember selection to prevent stale updates
        self._selected_key = key_str
        # Show selection info immediately
        self._update_selection_info(run)
        self.log(
            f'RowSelected: retrieving run code for {run.run_number}-{run.site_number}'
        )
        # Start background load; do not block the UI
        asyncio.create_task(self._load_and_display_run(run, key_str))

    async def _load_and_display_run(
        self, run: BocaRun, expected_key: Optional[str]
    ) -> None:
        try:
            # If already cached, use immediately
            cached = self._find_cached_run_path(run)
            if cached is not None:
                if expected_key is None or self._selected_key == expected_key:
                    self.code_box.path = cached
                    # Ensure selection info reflects final (not loading) state
                    self._update_selection_info(run, loading=False)
                else:
                    self.log(
                        f'_load_and_display_run: discard stale cached expected={expected_key} current={self._selected_key}'
                    )
                return

            # Otherwise, ensure cached in background and then display
            # Indicate loading on the right panel and in the code box
            if expected_key is None or self._selected_key == expected_key:
                try:
                    self._update_selection_info(run, loading=True)
                except Exception:
                    pass
            path = await self._ensure_cached_run(
                run, is_prefetch=False, priority=self.PRIORITY_USER
            )
            if path is None:
                return
            if expected_key is None or self._selected_key == expected_key:
                self.code_box.path = path
                # Clear loading indicator now that code is displayed
                self._update_selection_info(run, loading=False)
            else:
                self.log(
                    f'_load_and_display_run: discard stale result expected={expected_key} current={self._selected_key}'
                )
        except Exception:
            self.log('_load_and_display_run: error while retrieving or displaying run')

    @on(DataTable.RowHighlighted, '#runs_table')
    async def _on_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        # Track highlighted key for diff action
        self._highlighted_key = self._extract_key_str(event.row_key)
        await self._on_row_selected(event)  # pyright: ignore[reportArgumentType]
        self.log(
            f'RowHighlighted: event.row_key={event.row_key!r} -> highlighted={self._highlighted_key!r}'
        )

    def _write_temp_code(self, detailed: BocaDetailedRun) -> Path:
        filename = detailed.filename.name
        # Ensure unique file per run
        safe_name = f'{detailed.run_number}-{detailed.site_number}-{filename}'
        path = self._tmp_dir / safe_name
        path.write_text(detailed.code)
        self.log(f'_write_temp_code: wrote code to {path} (len={len(detailed.code)})')
        return path

    # --- Previous run helpers ---
    def _find_last_non_ac_before(self, run: BocaRun) -> Optional[BocaRun]:
        team = (run.user or '').strip()
        problem = run.problem_shortname
        site_number = run.site_number
        # Use pre-indexed list for (problem, team)
        indexed = self._runs_by_problem_user.get((problem, team), [])
        # List is sorted by (time, run_number); scan from the end for efficiency
        for prev in reversed(indexed):
            if prev.site_number != site_number:
                continue
            if prev.outcome is None or prev.outcome == Outcome.ACCEPTED:
                continue
            if (prev.time < run.time) or (
                prev.time == run.time and prev.run_number < run.run_number
            ):
                return prev
        return None

    # --- Request management & loading indicator ---
    async def _run_in_thread(
        self, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Any:
        self.pending_requests += 1
        try:
            return await asyncio.to_thread(func, *args, **kwargs)
        finally:
            self.pending_requests -= 1

    async def _run_scraper(
        self,
        func: Callable[..., Any],
        *args: Any,
        priority: int = 5,
        **kwargs: Any,
    ) -> Any:
        """Schedule a scraper call with priority; executes serially in a worker."""
        # Create a future to await result
        loop = asyncio.get_running_loop()
        fut: asyncio.Future = loop.create_future()
        # Use sequence to keep FIFO within same priority
        self._scraper_seq += 1
        item = (priority, self._scraper_seq, func, args, kwargs, fut)
        await self._scraper_queue.put(item)
        # Ensure worker is running
        self._start_scraper_worker()
        return await fut

    def _start_scraper_worker(self) -> None:
        if (
            self._scraper_worker_task is not None
            and not self._scraper_worker_task.done()
        ):
            return

        async def _worker() -> None:
            while True:
                try:
                    (
                        priority,
                        seq,
                        func,
                        args,
                        kwargs,
                        fut,
                    ) = await self._scraper_queue.get()  # type: ignore[misc]
                except asyncio.CancelledError:
                    break
                except Exception:
                    # Should not happen; continue loop
                    await asyncio.sleep(0)
                    continue
                try:
                    if not fut.cancelled():
                        result = await self._run_in_thread(func, *args, **kwargs)
                        fut.set_result(result)
                except asyncio.CancelledError:
                    if not fut.cancelled():
                        fut.set_exception(asyncio.CancelledError())
                    break
                except Exception as exc:
                    if not fut.cancelled():
                        fut.set_exception(exc)
                finally:
                    self._scraper_queue.task_done()

        self._scraper_worker_task = asyncio.create_task(_worker())

    def _cancel_scraper_worker(self) -> None:
        task = self._scraper_worker_task
        if task is not None and not task.done():
            task.cancel()
        self._scraper_worker_task = None

    def watch_pending_requests(self, value: int) -> None:  # called on change
        self._update_loading_indicator()

    def _update_loading_indicator(self) -> None:
        try:
            indicator = self.query_one('#loading_indicator', Static)
        except Exception:
            return
        prefetch_count = 0
        try:
            prefetch_count = len(self._prefetch_inflight)
        except Exception:
            prefetch_count = 0
        if self.pending_requests > 0:
            if prefetch_count > 0:
                indicator.update(f'⏳ Loading… (prefetch {prefetch_count})')
            else:
                indicator.update('⏳ Loading…')
        elif prefetch_count > 0:
            indicator.update(f'Prefetch in-flight: {prefetch_count}')
        else:
            indicator.update('')

    def on_unmount(self) -> None:
        # Ensure background tasks are stopped cleanly
        self._cancel_auto_refresh_task()
        self._cancel_scraper_worker()

    # --- Persistent cache helpers ---
    def _init_cache_paths(self) -> None:
        base = get_app_path() / 'boca' / 'contests' / (self.contest_id or 'default')
        runs_dir = base / 'runs'
        runs_dir.mkdir(parents=True, exist_ok=True)
        self._cache_base = base
        self._runs_dir = runs_dir

    def _run_key(self, run: BocaRun) -> str:
        return f'{run.run_number}:{run.site_number}'

    def _find_cached_run_path(self, run: BocaRun) -> Optional[Path]:
        if self._runs_dir is None:
            return None
        try:
            candidate = self._runs_dir / f'{run.run_number}-{run.site_number}'
            if candidate.is_file():
                return candidate
        except Exception:
            return None
        return None

    # --- Failed fetch sentinel helpers ---
    def _failure_sentinel_path(self, run: BocaRun) -> Optional[Path]:
        if self._runs_dir is None:
            return None
        return self._runs_dir / f'{run.run_number}-{run.site_number}.fail'

    def _read_failure_attempts(self, run: BocaRun) -> int:
        """Return number of failed fetch attempts recorded for this run."""
        try:
            path = self._failure_sentinel_path(run)
            if path is None or not path.is_file():
                return 0
            raw = path.read_text().strip()
            try:
                return max(0, int(raw))
            except Exception:
                # If contents are unexpected, treat as 1 failed attempt
                return 1
        except Exception:
            return 0

    def _record_fetch_failure(self, run: BocaRun) -> None:
        """Increment failure counter for this run (prefetch skip uses this)."""
        try:
            path = self._failure_sentinel_path(run)
            if path is None:
                return
            attempts = self._read_failure_attempts(run) + 1
            path.write_text(str(attempts))
        except Exception:
            # Best-effort only
            pass

    def _clear_fetch_failure(self, run: BocaRun) -> None:
        """Clear failure sentinel for this run after a successful fetch."""
        try:
            path = self._failure_sentinel_path(run)
            if path is not None and path.exists():
                path.unlink(missing_ok=True)  # type: ignore[call-arg]
        except Exception:
            # Ignore cleanup errors
            pass

    async def _ensure_cached_run(
        self, run: BocaRun, *, is_prefetch: bool, priority: int
    ) -> Optional[Path]:
        # Check existing
        existing = self._find_cached_run_path(run)
        if existing is not None:
            return existing
        key = self._run_key(run)
        # Track only prefetch tasks in the inflight set (for UI)
        if is_prefetch:
            self._prefetch_inflight.add(key)
            self._update_loading_indicator()
        try:
            detailed: BocaDetailedRun = await self._run_scraper(
                self.scraper.retrieve_run, run, priority=priority
            )
            if self._runs_dir is None:
                return None
            # Write code file with simplified name '<run>-<site>'
            base_name = f'{detailed.run_number}-{detailed.site_number}'
            path = self._runs_dir / base_name
            path.write_text(detailed.code)
            # Write metadata sidecar JSON with same stem
            meta = {
                'run_number': detailed.run_number,
                'site_number': detailed.site_number,
                'filename': str(getattr(detailed.filename, 'name', '')),
                'language': str(getattr(detailed, 'language_repr', '') or ''),
            }
            try:
                (self._runs_dir / f'{base_name}.json').write_text(
                    json.dumps(meta, ensure_ascii=False)
                )
            except Exception:
                # Metadata failures should not break caching
                pass
            self.log(
                f'_ensure_cached_run: cached code to {path} (len={len(detailed.code)})'
            )
            # Successful fetch -> clear any prior failure sentinel
            self._clear_fetch_failure(run)
            return path
        except Exception:
            # Only count failures that raise here
            self._record_fetch_failure(run)
            return None
        finally:
            if is_prefetch:
                self._prefetch_inflight.discard(key)
                self._update_loading_indicator()

    async def _prefetch_missing_runs(self) -> None:
        # Sequentially prefetch to honor scraper serialization
        for run in list(self._runs):
            try:
                if self._find_cached_run_path(run) is None:
                    # Skip prefetch if we've already failed 3 times for this run
                    if self._read_failure_attempts(run) >= 3:
                        continue
                    await self._ensure_cached_run(
                        run, is_prefetch=True, priority=self.PRIORITY_PREFETCH
                    )
            except Exception:
                # Ignore individual failures
                pass

    def _is_small_diff_done_for_run(self, run: BocaRun) -> bool:
        if run.outcome is None or run.outcome != Outcome.ACCEPTED:
            return False
        key = self._run_key(run)
        entry = self._small_diff_cache.get(key)
        if entry is None:
            return False
        try:
            return (
                int(entry.get('threshold', -1)) == int(self.small_diff_threshold)
                and entry.get('verdict') == run.outcome.name
            )
        except Exception:
            return False

    def _bg_status_for_run(self, run: BocaRun) -> str:
        has_p = self._find_cached_run_path(run) is not None
        has_d = self._is_small_diff_done_for_run(run)
        # Show 'X' when this run has reached the failure cap (>= 3 attempts)
        has_x = self._read_failure_attempts(run) >= 3
        parts: list[str] = []
        if has_p:
            parts.append('P')
        if has_d:
            parts.append('D')
        if has_x:
            parts.append('X')
        return ''.join(parts)

    async def _compute_small_diffs(self) -> None:
        """Compute and cache 'small diff' markers for AC runs.
        A run is marked if its diff to the last non-AC run for same team/problem/site
        has changed-line count <= small_diff_threshold.
        """
        threshold = int(self.small_diff_threshold)
        # Compute only for runs with missing/invalid cache
        for run in list(self._runs):
            try:
                if run.outcome is None or run.outcome != Outcome.ACCEPTED:
                    continue
                key = self._run_key(run)
                cached = self._small_diff_cache.get(key)
                verdict_name = run.outcome.name
                if cached is not None:
                    if (
                        int(cached.get('threshold', -1)) == threshold
                        and cached.get('verdict') == verdict_name
                    ):
                        # Cache is valid; skip recompute
                        continue
                # 1) If current submission has not been prefetched yet, skip for now
                path_current = self._find_cached_run_path(run)
                if path_current is None:
                    continue
                # 2) Ensure previous candidate is prefetched (fire request if needed)
                prev = self._find_last_non_ac_before(run)
                if prev is None:
                    # No previous non-AC to compare against
                    continue
                path_prev = await self._ensure_cached_run(
                    prev, is_prefetch=True, priority=self.PRIORITY_PREFETCH
                )
                if path_prev is None:
                    continue
                try:
                    a = path_prev.read_text().splitlines()
                    b = path_current.read_text().splitlines()
                except Exception:
                    continue
                diff_iter = difflib.unified_diff(a, b, lineterm='')
                changed = 0
                for line in diff_iter:
                    if not line:
                        continue
                    # Skip headers
                    if (
                        line.startswith('---')
                        or line.startswith('+++')
                        or line.startswith('@@')
                    ):
                        continue
                    if line[0] == '+' or line[0] == '-':
                        changed += 1
                        if changed > threshold:
                            break
                self._small_diff_cache[key] = {
                    'flag': changed <= threshold,
                    'changed': changed,
                    'threshold': threshold,
                    'verdict': verdict_name,
                }
            except Exception:
                # Ignore per-run compute failures
                pass
        # Update flags for current runs from valid cache
        flags: dict[str, bool] = {}
        for run in list(self._runs):
            try:
                if run.outcome is None or run.outcome != Outcome.ACCEPTED:
                    continue
                key = self._run_key(run)
                entry = self._small_diff_cache.get(key)
                if entry is None:
                    continue
                if (
                    int(entry.get('threshold', -1)) == threshold
                    and entry.get('verdict') == run.outcome.name
                ):
                    flags[key] = bool(entry.get('flag', False))
            except Exception:
                pass
        self._small_diff_flags = flags
        try:
            self._reload_table()
        except Exception:
            pass


def run_app(contest_id: Optional[str] = None) -> None:
    app = BocaRunsApp(contest_id=contest_id)
    app.run()


class _BocaDifferScreen(Screen):
    BINDINGS = [
        ('q', 'app.pop_screen', 'Quit'),
    ]

    def __init__(self, path1: Path, path2: Path):
        super().__init__()
        self._path1 = path1
        self._path2 = path2

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Vertical():
            yield DiffBox()

    def on_mount(self) -> None:
        diff = self.query_one(DiffBox)
        diff.paths = (self._path1, self._path2)


class _OptionsScreen(Screen):
    BINDINGS = [
        ('enter', 'save', 'Save'),
        ('escape', 'cancel', 'Cancel'),
        ('q', 'cancel', 'Quit'),
    ]

    def __init__(self, refresh_interval: int, small_diff_threshold: int):
        super().__init__()
        self._initial_refresh = int(refresh_interval)
        self._initial_diff = int(small_diff_threshold)

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Vertical():
            yield Static('Options: Set values and press Enter to save')
            yield Static('Auto refresh period (seconds):')
            yield Input(value=str(self._initial_refresh), id='opt_refresh_input')
            yield Static('Delta lines threshold (<= changes):')
            yield Input(value=str(self._initial_diff), id='opt_diff_input')

    def on_mount(self) -> None:
        try:
            inp = self.query_one('#opt_refresh_input', Input)
            self.set_focus(inp)
        except Exception:
            pass

    def _submit(self) -> None:
        try:
            refresh_inp = self.query_one('#opt_refresh_input', Input)
            diff_inp = self.query_one('#opt_diff_input', Input)
            refresh_raw = (refresh_inp.value or '').strip()
            diff_raw = (diff_inp.value or '').strip()
            self.log(
                f'_OptionsScreen._submit: raw values refresh={refresh_raw!r} diff={diff_raw!r}'
            )
        except Exception:
            self.log('_OptionsScreen._submit: failed to read inputs; dismissing None')
            self.dismiss(None)
            return
        # Fallback to initial values on invalid/empty
        try:
            refresh_val = int(refresh_raw) if refresh_raw else self._initial_refresh
            if refresh_val <= 0:
                refresh_val = self._initial_refresh
            self.log(f'_OptionsScreen._submit: parsed refresh={refresh_val}')
        except Exception:
            self.log(
                f'_OptionsScreen._submit: invalid refresh={refresh_raw!r}; using {self._initial_refresh}'
            )
            refresh_val = self._initial_refresh
        try:
            diff_val = int(diff_raw) if diff_raw else self._initial_diff
            if diff_val <= 0:
                diff_val = self._initial_diff
            self.log(f'_OptionsScreen._submit: parsed diff={diff_val}')
        except Exception:
            self.log(
                f'_OptionsScreen._submit: invalid diff={diff_raw!r}; using {self._initial_diff}'
            )
            diff_val = self._initial_diff
        payload = {'refresh_interval': refresh_val, 'small_diff_threshold': diff_val}
        self.log(f'_OptionsScreen._submit: dismissing with {payload!r}')
        self.dismiss(payload)

    @on(Input.Submitted, '#opt_refresh_input')
    def _on_submit_refresh(self, event: Input.Submitted) -> None:
        self._submit()

    @on(Input.Submitted, '#opt_diff_input')
    def _on_submit_diff(self, event: Input.Submitted) -> None:
        self._submit()

    def action_save(self) -> None:
        self._submit()

    def action_cancel(self) -> None:
        self.dismiss(None)



import pathlib
from unittest import mock

import pytest
import typer

from rbx.box import package, remote
from rbx.box.schema import ExpectedOutcome
from rbx.box.testing import testing_package
from rbx.box.tooling.boca.scraper import BocaRun


class TestExpander:
    """Test the base Expander class."""

    def test_get_remote_path(self, testing_pkg: testing_package.TestingPackage):
        """Test get_remote_path returns correct path under remote directory."""
        expander = remote.MainExpander()
        test_path = pathlib.Path('test/path')

        result = expander.get_remote_path(test_path)

        # Should be under the problem remote directory
        assert result.parent.parent.name == '.remote'
        assert result.name == 'path'

    def test_cacheable_paths_default_empty(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test cacheable_paths returns empty list by default."""
        expander = remote.MainExpander()

        result = expander.cacheable_paths(pathlib.Path('test'))

        assert result == []

    def test_cacheable_globs_default_empty(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test cacheable_globs returns empty list by default."""
        expander = remote.MainExpander()

        result = expander.cacheable_globs(pathlib.Path('test'))

        assert result == []


class TestMainExpander:
    """Test the MainExpander class."""

    def test_expand_with_main_path(self, testing_pkg: testing_package.TestingPackage):
        """Test expand returns main solution path when given @main."""
        # Set up main solution
        testing_pkg.add_solution('sols/main.cpp', outcome=ExpectedOutcome.ACCEPTED)

        expander = remote.MainExpander()

        result = expander.expand(pathlib.Path('@main'))

        assert result == pathlib.Path('sols/main.cpp')

    def test_expand_with_non_main_path(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand returns None when path is not @main."""
        expander = remote.MainExpander()

        result = expander.expand(pathlib.Path('@other'))

        assert result is None

    def test_expand_with_no_main_solution(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand returns None when no main solution exists."""
        # Add a non-accepted solution
        testing_pkg.add_solution('sols/wa.cpp', outcome=ExpectedOutcome.WRONG_ANSWER)

        expander = remote.MainExpander()

        result = expander.expand(pathlib.Path('@main'))

        assert result is None


class TestBocaExpander:
    """Test the BocaExpander class."""

    def test_get_match_valid_run_with_site(self):
        """Test get_match parses valid BOCA run with site number."""
        expander = remote.BocaExpander()

        result = expander.get_match('@boca/123-2')

        assert result == (123, 2)

    def test_get_match_valid_run_without_site(self):
        """Test get_match parses valid BOCA run without site number (defaults to 1)."""
        expander = remote.BocaExpander()

        result = expander.get_match('@boca/456')

        assert result == (456, 1)

    def test_get_match_invalid_format(self):
        """Test get_match returns None for invalid format."""
        expander = remote.BocaExpander()

        result = expander.get_match('@invalid/format')

        assert result is None

    def test_get_match_non_boca_path(self):
        """Test get_match returns None for non-BOCA path."""
        expander = remote.BocaExpander()

        result = expander.get_match('@main')

        assert result is None

    def test_get_boca_path(self, testing_pkg: testing_package.TestingPackage):
        """Test get_boca_path returns correct path for run and site."""
        expander = remote.BocaExpander()

        result = expander.get_boca_path(123, 2)

        assert result.name == '123-2'
        assert result.parent.name == 'boca'

    def test_cacheable_globs_valid_boca_path(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test cacheable_globs returns correct glob pattern for valid BOCA path."""
        expander = remote.BocaExpander()

        result = expander.cacheable_globs(pathlib.Path('@boca/123-2'))

        assert len(result) == 1
        assert result[0].endswith('123-2.*')
        assert 'boca' in result[0]

    def test_cacheable_globs_invalid_path(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test cacheable_globs returns empty list for invalid path."""
        expander = remote.BocaExpander()

        result = expander.cacheable_globs(pathlib.Path('@main'))

        assert result == []

    @mock.patch('rbx.box.tooling.boca.scraper.get_boca_scraper')
    def test_expand_valid_boca_path(
        self, mock_get_scraper, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand downloads and returns path for valid BOCA run."""
        # Mock the BOCA scraper
        mock_scraper = mock.MagicMock()
        mock_get_scraper.return_value = mock_scraper

        # Create a temporary file to simulate downloaded solution
        boca_folder = package.get_problem_remote_dir() / 'boca'
        boca_folder.mkdir(parents=True, exist_ok=True)
        downloaded_file = boca_folder / '123-2.cpp'

        mock_scraper.download_run.return_value = downloaded_file

        expander = remote.BocaExpander()

        result = expander.expand(pathlib.Path('@boca/123-2'))

        # Verify scraper was called correctly
        mock_scraper.login.assert_called_once()
        mock_scraper.download_run.assert_called_once_with(
            BocaRun.from_run_number(123, 2), expander.get_boca_folder()
        )

        assert result == downloaded_file

    def test_expand_invalid_boca_path(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand returns None for invalid BOCA path."""
        expander = remote.BocaExpander()

        result = expander.expand(pathlib.Path('@main'))

        assert result is None


class TestExpandFiles:
    """Test the expand_files function."""

    def test_expand_files_normal_paths(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_files passes through normal paths unchanged."""
        files = ['normal.cpp', 'another.cpp']

        result = remote.expand_files(files)

        assert result == [pathlib.Path('normal.cpp'), pathlib.Path('another.cpp')]

    def test_expand_files_with_main_expansion(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_files expands @main to main solution."""
        # Set up main solution
        testing_pkg.add_solution('sols/main.cpp', outcome=ExpectedOutcome.ACCEPTED)

        files = ['normal.cpp', '@main']

        result = remote.expand_files(files)

        assert result == [pathlib.Path('normal.cpp'), pathlib.Path('sols/main.cpp')]

    def test_expand_files_with_unexpandable_paths(
        self, testing_pkg: testing_package.TestingPackage, capsys
    ):
        """Test expand_files skips unexpandable paths and prints warning."""
        files = ['normal.cpp', '@unknown']

        result = remote.expand_files(files)

        # Should skip the unexpandable path
        assert result == [pathlib.Path('normal.cpp')]

        # Should print a warning
        captured = capsys.readouterr()
        assert 'could not be expanded' in captured.out

    def test_expand_files_with_cached_file(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_files uses cached file when available."""
        # Create a cached file for BOCA expansion
        remote_dir = package.get_problem_remote_dir()
        boca_folder = remote_dir / 'boca'
        boca_folder.mkdir(parents=True, exist_ok=True)
        cached_file = boca_folder / '123-2.cpp'
        cached_file.write_text('cached content')

        files = ['@boca/123-2']

        result = remote.expand_files(files)

        # Should return the cached file path relative to package
        assert result == [testing_pkg.relpath(cached_file)]
        assert cached_file.read_text() == 'cached content'

    def test_expand_files_mixed_paths(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_files handles mixed normal and remote paths."""
        # Set up main solution
        testing_pkg.add_solution('sols/main.cpp', outcome=ExpectedOutcome.ACCEPTED)

        files = ['normal1.cpp', '@main', 'normal2.cpp']

        result = remote.expand_files(files)

        assert result == [
            pathlib.Path('normal1.cpp'),
            pathlib.Path('sols/main.cpp'),
            pathlib.Path('normal2.cpp'),
        ]

    @mock.patch('rbx.box.cd.is_problem_package')
    def test_expand_files_not_in_package(
        self, mock_is_problem_package, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_files exits when not in a problem package."""
        mock_is_problem_package.return_value = False

        with pytest.raises(typer.Exit):
            remote.expand_files(['@main'])


class TestExpandFile:
    """Test the expand_file function."""

    def test_expand_file_single_result(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_file returns single expanded file."""
        # Set up main solution
        testing_pkg.add_solution('sols/main.cpp', outcome=ExpectedOutcome.ACCEPTED)

        result = remote.expand_file('@main')

        assert result == pathlib.Path('sols/main.cpp')

    def test_expand_file_no_expansion(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_file exits when expansion fails."""
        with pytest.raises(typer.Exit):
            remote.expand_file('@unknown')

    def test_expand_file_multiple_results(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test expand_file exits when multiple results returned (shouldn't happen in practice)."""
        # This is a bit artificial since expand_files normally returns one result per input
        # but we test the error handling
        with mock.patch.object(remote, 'expand_files') as mock_expand:
            mock_expand.return_value = [
                pathlib.Path('file1.cpp'),
                pathlib.Path('file2.cpp'),
            ]

            with pytest.raises(typer.Exit):
                remote.expand_file('@test')


class TestIsPathRemote:
    """Test the is_path_remote function."""

    def test_is_path_remote_true(self, testing_pkg: testing_package.TestingPackage):
        """Test is_path_remote returns True for paths under remote directory."""
        remote_dir = package.get_problem_remote_dir()
        remote_file = remote_dir / 'test.cpp'
        remote_file.parent.mkdir(parents=True, exist_ok=True)
        remote_file.write_text('test content')

        result = remote.is_path_remote(remote_file)

        assert result is True

    def test_is_path_remote_false(self, testing_pkg: testing_package.TestingPackage):
        """Test is_path_remote returns False for paths outside remote directory."""
        local_file = testing_pkg.path('local.cpp')
        local_file.write_text('local content')

        result = remote.is_path_remote(local_file)

        assert result is False

    def test_is_path_remote_relative_path(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test is_path_remote works with relative paths."""
        # Create a file in the remote directory
        remote_dir = package.get_problem_remote_dir()
        remote_file = remote_dir / 'test.cpp'
        remote_file.parent.mkdir(parents=True, exist_ok=True)
        remote_file.write_text('test content')

        # Test with the actual relative path to the remote file
        result = remote.is_path_remote(remote_file)

        assert result is True


class TestBocaRegex:
    """Test the BOCA_REGEX pattern."""

    def test_boca_regex_with_site(self):
        """Test BOCA_REGEX matches pattern with site number."""
        match = remote.BocaExpander.BOCA_REGEX.match('@boca/123-2')

        assert match is not None
        assert match.group(1) == '123'
        assert match.group(2) == '2'

    def test_boca_regex_without_site(self):
        """Test BOCA_REGEX matches pattern without site number."""
        match = remote.BocaExpander.BOCA_REGEX.match('@boca/456')

        assert match is not None
        assert match.group(1) == '456'
        assert match.group(2) is None

    def test_boca_regex_invalid_format(self):
        """Test BOCA_REGEX doesn't match invalid format."""
        match = remote.BocaExpander.BOCA_REGEX.match('@boca/invalid')

        assert match is None

    def test_boca_regex_non_boca_path(self):
        """Test BOCA_REGEX doesn't match non-BOCA paths."""
        match = remote.BocaExpander.BOCA_REGEX.match('@main')

        assert match is None



import pathlib
import shutil
from math import fabs
from typing import List, Optional

import typer

from rbx import console
from rbx.box import header, limits_info, naming, package
from rbx.box.environment import get_extension_or_default
from rbx.box.packaging.boca.extension import BocaExtension, BocaLanguage
from rbx.box.packaging.packager import BasePackager, BuiltStatement
from rbx.box.schema import TaskType
from rbx.box.statements.schema import Statement
from rbx.config import get_default_app_path, get_testlib

_MAX_REP_TIME = (
    7  # TL to allow for additional rounding reps should be < _MAX_REP_TIME in seconds
)
_MAX_REPS = 10  # Maximum number of reps to add


def test_time(time):
    return max(1, round(time))


class BocaPackager(BasePackager):
    def __init__(self, language: Optional[str] = None):
        self.language = language

    def languages(self) -> List[str]:
        if self.language is None:
            return super().languages()
        return [self.language]

    @classmethod
    def task_types(cls) -> List[TaskType]:
        return [TaskType.BATCH, TaskType.COMMUNICATION]

    def _get_main_statement(self) -> Optional[Statement]:
        pkg = package.find_problem_package_or_die()

        if not pkg.expanded_statements:
            return None

        if self.language is None:
            return pkg.expanded_statements[0]

        for statement in pkg.expanded_statements:
            if statement.language == self.language:
                return statement

        return None

    def _get_main_built_statement(
        self, built_statements: List[BuiltStatement]
    ) -> Optional[BuiltStatement]:
        statement = self._get_main_statement()
        if statement is None:
            return None

        for built_statement in built_statements:
            if built_statement.statement == statement:
                return built_statement

        return None

    def _get_problem_name(self) -> str:
        # BOCA forces Java class names to be the name of the problem.
        return self.package_basename().replace('-', '_')

    def _get_zip_filename(self) -> str:
        if self.language is None:
            return self._get_problem_name()
        return f'{self._get_problem_name()}-{self.language}'

    def _get_problem_basename(self) -> str:
        extension = get_extension_or_default('boca', BocaExtension)
        shortname = naming.get_problem_shortname()
        if extension.preferContestLetter and shortname is not None:
            return shortname
        return self._get_problem_name()

    def _get_problem_info(self) -> str:
        statement = self._get_main_statement()
        lang = statement.language if statement is not None else None
        title = naming.get_problem_title(lang, statement, fallback_to_title=True)
        return (
            f'basename={self._get_problem_basename()}\n'
            f'fullname={title}\n'
            f'descfile={self._get_problem_basename()}.pdf\n'
        )

    def _get_pkg_timelimit(self, language: BocaLanguage) -> int:
        limits = limits_info.get_limits(language, profile='boca')
        assert limits.time is not None
        return limits.time

    def _get_pkg_memorylimit(self, language: BocaLanguage) -> int:
        limits = limits_info.get_limits(language, profile='boca')
        assert limits.memory is not None
        return limits.memory

    def _get_number_of_runs(self, language: BocaLanguage) -> int:
        pkg = package.find_problem_package_or_die()
        extension = get_extension_or_default('boca', BocaExtension)
        pkg_timelimit = self._get_pkg_timelimit(language)
        time = pkg_timelimit / 1000  # convert to seconds

        if time >= _MAX_REP_TIME:
            console.console.print(
                f'[warning]Use time limit of {time} seconds instead of {pkg_timelimit}ms because TL is too large.[/warning]'
            )
            return 1

        def rounding_error(time):
            return fabs(time - test_time(time))

        def error_percentage(time, runs):
            return rounding_error(time * runs) / (time * runs)

        for i in range(1, _MAX_REPS + 1):
            if error_percentage(time, i) <= extension.maximumTimeError:
                console.console.print(
                    f'[warning]Using {i} run(s) to define integer TL for BOCA when using language [item]{language}[/item] '
                    f'(original TL is {pkg_timelimit}ms, new TL is {test_time(time * i) * 1000}ms).[/warning]'
                )
                return i

        percent_str = f'{round(extension.maximumTimeError * 100)}%'
        console.console.print(
            f'[error]Error while defining limits for problem [item]{pkg.name}[/item], language [item]{language}[/item].[/error]'
        )
        console.console.print(
            f'[error]Introducing an error of less than {percent_str} in the TL in less than '
            f'{_MAX_REPS} runs is not possible.[/error]'
        )
        console.console.print(
            f'[error]Original TL for [item]{language}[/item] is {pkg_timelimit}ms, please review it.[/error]'
        )
        raise typer.Exit(1)

    def _get_limits(self, language: BocaLanguage) -> str:
        pkg = package.find_problem_package_or_die()
        if pkg.type == TaskType.COMMUNICATION:
            # Interactive tasks only support a single run.
            no_of_runs = 1
            time_limit = f'{self._get_pkg_timelimit(language) / 1000:.2f}'
        else:
            no_of_runs = self._get_number_of_runs(language)
            time_limit = test_time(
                self._get_pkg_timelimit(language) / 1000 * no_of_runs
            )
        return (
            '#!/bin/bash\n'
            f'echo {time_limit}\n'
            f'echo {no_of_runs}\n'
            f'echo {self._get_pkg_memorylimit(language)}\n'
            f'echo {pkg.outputLimit}\n'
            f'exit 0\n'
        )

    def _get_compare(self) -> str:
        compare_path = get_default_app_path() / 'packagers' / 'boca' / 'compare.sh'
        if not compare_path.exists():
            console.console.print(
                '[error]BOCA template compare script not found.[/error]'
            )
            raise typer.Exit(1)
        return compare_path.read_text()

    def _replace_common(self, text: str, lang: str) -> str:
        extension = get_extension_or_default('boca', BocaExtension)
        flags = extension.flags_with_defaults()
        if lang in flags:
            text = text.replace('{{rbxFlags}}', flags[lang])
        return text.replace(
            '{{rbxPython3}}', 'pypy3' if extension.usePypy else 'python3'
        )

    def _get_checker(self) -> str:
        checker_path = get_default_app_path() / 'packagers' / 'boca' / 'checker.sh'
        if not checker_path.exists():
            console.console.print(
                '[error]BOCA template checker script not found.[/error]'
            )
            raise typer.Exit(1)
        checker_text = checker_path.read_text()
        testlib = get_testlib().read_text()
        checker = package.get_checker_or_builtin().path.read_text()
        rbx_header = header.get_header().read_text()
        return (
            self._replace_common(checker_text, 'cc')
            .replace('{{testlib_content}}', testlib)
            .replace('{{rbx_header_content}}', rbx_header)
            .replace('{{checker_content}}', checker)
        )

    def _get_interactor(self) -> str:
        interactor_path = (
            get_default_app_path() / 'packagers' / 'boca' / 'interactor_compile.sh'
        )
        if not interactor_path.exists():
            console.console.print(
                '[error]BOCA template interactor compile script not found.[/error]'
            )
            raise typer.Exit(1)

        interactor_text = interactor_path.read_text()
        interactor = package.get_interactor().path.read_text()
        return self._replace_common(interactor_text, 'cc').replace(
            '{{interactor_content}}', interactor
        )

    def _get_safeexec(self) -> str:
        safeexec_script_path = (
            get_default_app_path() / 'packagers' / 'boca' / 'safeexec_compile.sh'
        )
        safeexec_path = get_default_app_path() / 'packagers' / 'boca' / 'safeexec.c'
        if not safeexec_script_path.exists():
            console.console.print(
                '[error]BOCA template safeexec compile script not found.[/error]'
            )
            raise typer.Exit(1)
        if not safeexec_path.exists():
            console.console.print(
                '[error]BOCA template safeexec source code not found.[/error]'
            )
            raise typer.Exit(1)
        return safeexec_script_path.read_text().replace(
            '{{safeexec_content}}', safeexec_path.read_text()
        )

    def _get_pipe(self) -> str:
        pipe_script_path = (
            get_default_app_path() / 'packagers' / 'boca' / 'pipe_compile.sh'
        )
        pipe_path = get_default_app_path() / 'packagers' / 'boca' / 'pipe.c'
        if not pipe_script_path.exists():
            console.console.print(
                '[error]BOCA template pipe compile script not found.[/error]'
            )
            raise typer.Exit(1)
        if not pipe_path.exists():
            console.console.print(
                '[error]BOCA template pipe source code not found.[/error]'
            )
            raise typer.Exit(1)
        return pipe_script_path.read_text().replace(
            '{{pipe_content}}', pipe_path.read_text()
        )

    def _get_compile(self, language: BocaLanguage) -> str:
        pkg = package.find_problem_package_or_die()

        compile_path = (
            get_default_app_path() / 'packagers' / 'boca' / 'compile' / language
        )
        if not compile_path.is_file():
            console.console.print(
                f'[error]Compile script for language [item]{language}[/item] not found.[/error]'
            )
            raise typer.Exit(1)

        compile_text = compile_path.read_text()

        assert 'umask 0022' in compile_text
        if pkg.type == TaskType.COMMUNICATION:
            compile_text = compile_text.replace(
                'umask 0022', 'umask 0022\n\n' + self._get_interactor()
            )
            compile_text = compile_text.replace(
                'umask 0022', 'umask 0022\n\n' + self._get_safeexec()
            )
            compile_text = compile_text.replace(
                'umask 0022', 'umask 0022\n\n' + self._get_pipe()
            )
        compile_text = compile_text.replace(
            'umask 0022', 'umask 0022\n\n' + self._get_checker()
        )

        compile_text = self._replace_common(compile_text, language)
        return compile_text

    def _copy_solutions(self, into_path: pathlib.Path):
        for solution in package.get_solutions():
            dest_path = (
                into_path
                / solution.path.stem
                / pathlib.Path(self._get_problem_name()).with_suffix(
                    solution.path.suffix
                )
            )
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(str(solution.path), dest_path)

    def _expand_run_script(self, run_path: pathlib.Path, language: BocaLanguage):
        pkg = package.find_problem_package_or_die()
        content = run_path.read_text()
        if pkg.type == TaskType.COMMUNICATION:
            runit_content = (
                get_default_app_path() / 'packagers' / 'boca' / 'interactor_run.sh'
            ).read_text()
            content = content.replace('{{runit_content}}', runit_content)

        content = self._replace_common(content, language)
        run_path.write_text(content)

    @classmethod
    def name(cls) -> str:
        return 'boca'

    def package(
        self,
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltStatement],
    ) -> pathlib.Path:
        extension = get_extension_or_default('boca', BocaExtension)
        pkg = package.find_problem_package_or_die()
        # Prepare limits
        if 'boca' not in limits_info.get_available_profile_names():
            console.console.print(
                '[error]Required limits profile [item]boca[/item] not found.[/error]'
            )
            console.console.print(
                '[error]Make sure to run [item]rbx time -p boca[/item] to create the limits profile.[/error]'
            )
            raise typer.Exit(1)
        limits_path = into_path / 'limits'
        limits_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (limits_path / language).write_text(self._get_limits(language))

        # Prepare compare
        compare_path = into_path / 'compare'
        compare_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (compare_path / language).write_text(self._get_compare())

        # Prepare run
        run_path = into_path / 'run'
        run_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            run_orig_path = (
                get_default_app_path() / 'packagers' / 'boca' / 'run' / language
            )
            if pkg.type == TaskType.COMMUNICATION:
                run_orig_path = (
                    get_default_app_path()
                    / 'packagers'
                    / 'boca'
                    / 'interactive'
                    / language
                )
            if not run_orig_path.is_file():
                console.console.print(
                    f'[error]Run script for language [item]{language}[/item] not found for task of type [item]{pkg.type}[/item].[/error]'
                )
                raise typer.Exit(1)
            shutil.copyfile(run_orig_path, run_path / language)
            self._expand_run_script(run_path / language, language)

        # Prepare compile.
        compile_path = into_path / 'compile'
        compile_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (compile_path / language).write_text(self._get_compile(language))

        # Prepare tests
        tests_path = into_path / 'tests'
        tests_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (tests_path / language).write_text('exit 0\n')

        # Problem statement
        main_built_statement = self._get_main_built_statement(built_statements)
        description_path = into_path / 'description'
        description_path.mkdir(parents=True, exist_ok=True)
        (description_path / 'problem.info').write_text(self._get_problem_info())
        pdf_path = (description_path / self._get_problem_basename()).with_suffix('.pdf')
        if main_built_statement is not None:
            shutil.copyfile(
                main_built_statement.path,
                pdf_path,
            )
        else:
            pdf_path.touch(exist_ok=True)

        # Copy solutions
        solutions_path = into_path / 'solutions'
        solutions_path.mkdir(parents=True, exist_ok=True)
        self._copy_solutions(solutions_path)

        # Prepare IO
        inputs_path = into_path / 'input'
        inputs_path.mkdir(parents=True, exist_ok=True)
        outputs_path = into_path / 'output'
        outputs_path.mkdir(parents=True, exist_ok=True)

        testcases = self.get_flattened_built_testcases()
        for i, testcase in enumerate(testcases):
            shutil.copyfile(testcase.inputPath, inputs_path / f'{i + 1:03d}')
            if testcase.outputPath is not None:
                shutil.copyfile(testcase.outputPath, outputs_path / f'{i + 1:03d}')
            else:
                (outputs_path / f'{i + 1:03d}').touch()

        # Zip all.
        shutil.make_archive(
            str(build_path / self._get_zip_filename()), 'zip', into_path
        )

        return (build_path / self._get_zip_filename()).with_suffix('.zip')
